/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

/**
* @file_name kbp.c
*/

/**
* @defgroup kbp Knowledge Based Processor (KBP) API
* @brief Description about the KBP apis supported in Hyperexec
*/


#include <stdio.h>
#include <stdlib.h>


#include <sys/mman.h>
#include <stdint.h>
#include <sys/stat.h>
#include <unistd.h>
#include <mempool.h>
#include <sys/nlm_dev.h>

#include "nlm_hal_xlp_dev.h"
#include "nlm_hal_fmn.h"
#include "nlm_hal_kbp.h"


#define XLP_ILA_MAJOR 	260
#define XLP_ILA_NAME "nlm_xlp_ila"

#define CACHELINE_SIZE          (1ULL << 6)
#define CACHELINE_ALIGNED_ADDR(addr) (((unsigned long)(addr)) & ~(CACHELINE_SIZE-1))

#define ILA_STATION_ID	260
#define ILA_CODE	0xf


#define KBP_OK 0
#define KBP_FAIL -1


#define NLM_KBP_PRINT_INFO        1


#ifdef NLM_KBP_PRINT_INFO
#define kbp_print_info     printf
#else
#define kbp_print_info(...)
#endif

#ifdef NLM_KBP_PRINT_DEBUG
#define kbp_print_debug     printf
#else
#define kbp_print_debug(...)
#endif


#define KBP_MEM_PKT_SIZE (12*8)
#define KBP_NUM_OF_PKT 4
#define KBP_NUM_OF_PKT_PER_VC 2


#define KBP_PKT_ALIGN_LEN	CACHELINE_SIZE


typedef struct kbp_mempool_data_t
{
	mpoolptr_t m_kbp_mempool;

	uint8_t m_is_pool_valid;
	uint8_t m_kbp_pkt_inuse[KBP_NUM_OF_PKT];
	uint8_t *m_kbp_pkt_mem[KBP_NUM_OF_PKT];
} kbp_mempool_data_t;


static kbp_mempool_data_t g_kbp_mempool;


#ifdef NLM_KBP_PRINT_DEBUG
static void
kbp_pvt_dump_ila_reg ()
{
	int32_t i = 0;
	printf ("\n\n dump interlaken LA register\n");
	for (i = 0; i < 6; i++)
		kbp_print_debug (" i: 0x%x val: 0x%x\n", i,
			 cam_reg_read (0, KBP_START_LANE, i));

	for (i = 0x30; i < 0x38; i++)
		kbp_print_debug (" i: 0x%x val: 0x%x\n", i,
			 cam_reg_read (0, KBP_START_LANE, i));

	for (i = 0x3c; i < 0x3F; i++)
		kbp_print_debug (" i: 0x%x val: 0x%x\n", i,
			 cam_reg_read (0, KBP_START_LANE, i));

	kbp_print_debug ("CAM Reg\n");
	for (i = 0x40; i < 0x55; i++)
		kbp_print_debug (" i: 0x%x val: 0x%x\n", i,
			 cam_reg_read (0, KBP_START_LANE, i));

	kbp_print_debug ("InterLA Reg\n");
	for (i = 0x55; i < 0x61; i++)
		kbp_print_debug (" i: 0x%x val: 0x%x\n", i,
			 cam_reg_read (0, KBP_START_LANE, i));

	kbp_print_debug ("SerDes Reg\n");
	for (i = 0x61; i < 0x65; i++)
		kbp_print_debug (" i: 0x%x val: 0x%x\n", i,
			 cam_reg_read (0, KBP_START_LANE, i));

	kbp_print_debug ("Memory Bar \n");
	for (i = 0x68; i < 0x6A; i++)
		kbp_print_debug (" i: 0x%x val: 0x%x\n", i,
			 cam_reg_read (0, KBP_START_LANE, i));
}
#endif /* NLM_KBP_PRINT_DEBUG */


static int
kbp_pvt_init_mempool (kbp_mempool_data_t * kbp_mempool)
{
	int32_t i = 0;

	kbp_print_info ("\n\n>>>> Initializing KBP memory pool.\n");

	if (kbp_mempool->m_is_pool_valid)
	{
		kbp_print_info ("\n\n>>>> KBP memory pool is already initialized.\n");
		return KBP_OK;
	}

	kbp_mempool->m_kbp_mempool = mpcreate ("kbp-app-pkt-pool",
					 KBP_NUM_OF_PKT,
					 KBP_MEM_PKT_SIZE,
					 KBP_PKT_ALIGN_LEN, 0);

	if (kbp_mempool->m_kbp_mempool == NULL)
	{
		kbp_print_info ("Failed to Allocate KBP memory pool\n");
		return KBP_FAIL;
	}

	for (i = 0; i < KBP_NUM_OF_PKT; ++i)
	{
		kbp_mempool->m_kbp_pkt_mem[i] = mpalloc (kbp_mempool->m_kbp_mempool, 0);
		if (kbp_mempool->m_kbp_pkt_mem[i] == NULL)
		{
			kbp_print_info ("Failed to allocate KBP memory for pkt #%d\n",
						i);
			return KBP_FAIL;
		}
		kbp_mempool->m_kbp_pkt_inuse[i] = 0;
	}

	kbp_mempool->m_is_pool_valid = 1;

	return KBP_OK;

}

/* Get the buffer from the pool for a given vc.
 We are keeping seperate buffer for each VC.
 Supporting only two VCs - even and odd as of now. */
static inline uint8_t *
kbp_pvt_get_buffer (kbp_mempool_data_t * kbp_mempool, int32_t vc)
{
	int32_t i = 0;

	vc &= 1;			/* Supporting only two VCs - even and odd as of now */

	for (i = vc * KBP_NUM_OF_PKT_PER_VC;
		i < (vc + 1) * KBP_NUM_OF_PKT_PER_VC; i++)
	{
		if (kbp_mempool->m_kbp_pkt_inuse[i] == 0)
		{
			kbp_mempool->m_kbp_pkt_inuse[i] = 1;
			return kbp_mempool->m_kbp_pkt_mem[i];
		}
	}
	return NULL;
}

/* Return the buffer to the pool*/
static inline void
kbp_pvt_return_buffer (kbp_mempool_data_t * kbp_mempool,
			   uint8_t * kbp_buff)
{
	int32_t i = 0;

	for (i = 0; i < KBP_NUM_OF_PKT; ++i)
	{
		if ((kbp_mempool->m_kbp_pkt_inuse[i] == 1) &&
			(kbp_buff == kbp_mempool->m_kbp_pkt_mem[i]))
		{
			kbp_mempool->m_kbp_pkt_inuse[i] = 0;
			return;
		}
	}
	return;
}

/* Destory memory pool*/
static inline void
kbp_pvt_destroy_mempool (kbp_mempool_data_t * kbp_mempool)
{
	int32_t i = 0;

	kbp_print_info ("\n\n>>>> Destroying KBP memory pool.\n");

	if (!kbp_mempool->m_is_pool_valid)
	{
		kbp_print_info ("\n\n>>>> KBP memory pool is already destroyed.\n");
		return;
	}

	for (i = 0; i < KBP_NUM_OF_PKT; ++i)
	{
		mpfree (kbp_mempool->m_kbp_mempool, kbp_mempool->m_kbp_pkt_mem[i]);
		kbp_mempool->m_kbp_pkt_inuse[i] = 0;
		kbp_mempool->m_kbp_pkt_mem[i] = NULL;
	}

	mpdestroy (kbp_mempool->m_kbp_mempool);
	kbp_mempool->m_kbp_mempool = NULL;
	kbp_mempool->m_is_pool_valid = 0;

	return;

}

/* Format the data and send the memory request to kbp*/
static inline int32_t
kbp_pvt_send_mem_req (uint32_t vc, kbp_rqt_data_t * req_p)
{
	unsigned int cam_msg_vc_base = ILA_STATION_ID;
	uint64_t xlp_req[4] = { 0x0ULL, };
	uint64_t phys_add;
	uint8_t *char_buff = NULL;
	int32_t ret_val = KBP_OK;
	uint8_t i = 0;

	char_buff = kbp_pvt_get_buffer (&g_kbp_mempool, vc);
	if (char_buff == NULL)
	{
		kbp_print_info ("!!!!Failed to allocate KBP memory !!!!");
		kbp_print_info (" VC #%d, request header = %x\n", vc,
		req_p->req_header);
		return KBP_FAIL;
	}

	memcpy (char_buff,
		(unsigned char *) (req_p->req_data), (req_p->valid_ln + 1) * 8);

	phys_add = mpgetphys (g_kbp_mempool.m_kbp_mempool,
			(void *) char_buff) & 0xffffffffc0ULL;

#ifdef NLM_KBP_PRINT_DEBUG
	kbp_print_debug ("Phy Add = %x, Req Add = %x\n",
				char_buff, req_p->req_data[0]);

	for (i = 0; i <= req_p->valid_ln; i++)
		kbp_print_debug ("Req Data [%d]= %lx\n", i, req_p->req_data[i]);

	for (i = 0; i <= req_p->valid_ln; i++)
	{
		kbp_print_debug ("\nMem Data [%d]= ", i);
		for (j = 0; j < 8; j++)
			kbp_print_debug ("%x ", char_buff[i * 8 + j]);
	}

#endif

	i = (req_p->req_len << 4) + req_p->empty_byte;
	xlp_req[0] = cam_mem_req_desc (i, phys_add);


	kbp_print_debug
		("\n\nData send to KBP = <%lx %lx %lx %lx> len = %d\n\n", xlp_req[0],
		xlp_req[1], xlp_req[2], xlp_req[3], req_p->valid_ln);


	ret_val = xlp_message_send (cam_msg_vc_base, 1, ILA_CODE, xlp_req);

	/* Return the buffer here itself. Later we need to place it inside kbp_recv() (NKG) */
	kbp_pvt_return_buffer (&g_kbp_mempool, char_buff);

	return ret_val;
}

/* Format the data and send direct request to kbp*/
static inline int
kbp_pvt_send_direct_req_direct_rsp (uint32_t vc, kbp_rqt_data_t * req_p)
{
	unsigned int cam_msg_vc_base = ILA_STATION_ID;
	uint64_t xlp_req[4] = { 0x0ULL, };
	uint8_t cam_rsp_type = 0;

	xlp_req[0] = cam_direct_req_desc (req_p->valid_ln,
				    cam_rsp_type, req_p->rsp_len,
				    req_p->req_len, req_p->empty_byte,
				    vc, req_p->req_header);

	switch (req_p->valid_ln)
	{
		case 3:
			xlp_req[3] = req_p->req_data[2];
		case 2:
			xlp_req[2] = req_p->req_data[1];
		case 1:
			xlp_req[1] = req_p->req_data[0];
	}

	kbp_print_debug
		("\n\nData send to KBP = <%lx %lx %lx %lx> len = %d\n\n", xlp_req[0],
		xlp_req[1], xlp_req[2], xlp_req[3], req_p->valid_ln);

	if (xlp_message_send
		(cam_msg_vc_base, req_p->valid_ln + 1, ILA_CODE, xlp_req))
	{
		return KBP_FAIL;
	}

	return KBP_OK;
}

/* Initialize the kbp link   */
static int32_t
kbp_ila_init_module (void)
{
	int32_t ret_val = KBP_OK;

#ifdef NLM_KBP_PRINT_DEBUG
	kbp_pvt_dump_ila_reg ();
#endif

	kbp_print_info ("nlm_hal_ila_serdes_init\n");
	ret_val = nlm_hal_ila_serdes_init ();

	if (ret_val != KBP_OK)
		return ret_val;

	kbp_print_info ("nlm_hal_ila_link_init\n");
	ret_val = nlm_hal_ila_link_init ();

	return ret_val;
}

/* Error handling in kbp operations  */
int32_t
kbp_handle_error (uint32_t vc, kbp_rsp_data_t * recv_rsp)
{
#ifdef NEED_TO_CHECK
	kbp_print_info ("Recieved Error Response: VC = %d, req id = %d\n",
				vc, recv_rsp->req_id);
#endif
	return KBP_OK;
}


/******************** KBP API ****************************/
/**
* @file_name kbp.c
*/

/**
* @defgroup kbp Knowledge Based Processor (KBP) API
* @brief Description about the KBP apis supported in Hyperexec
*/


/**
* @brief: kbp_init function is used to initialize the kbp link and other
*		 data structures required.
*
* @param[in]  vc_id		:virtual channle id
* @param[in]  init_link   :specifies whether to initialize link or not
*
* @return
*  - On sucess,  returns 0
*  - On failure, returns -1
*
* @sa kbp_init
* @ingroup kbp
*
*/
int32_t
kbp_init (uint32_t vc_id, uint32_t init_link)
{
	int32_t ret_val = KBP_OK;

	if (init_link)
	{
		kbp_print_info ("Calling link up now \n");
		ret_val = kbp_ila_init_module ();
		if (ret_val == KBP_OK)
		{
			kbp_print_info ("Done!! Exiting link up now \n");
		}
		else
		{
			kbp_print_info ("Error while link up...Please Try again \n");
			return ret_val;
		}
	}

	ret_val = kbp_pvt_init_mempool (&g_kbp_mempool);

	return ret_val;
}


/**
* @brief: kbp_send function is used to send the request to the
*		  kbp in fmn format.
*
* @param[in]  vc_id     :virtual channle id
* @param[in]  req_p    :pointer to kbp request data structure
*
* @return
*  - On sucess,  returns 0
*  - On failure, returns -1
*
* @sa kbp_send
* @ingroup kbp
*
*/
int32_t
kbp_send (uint32_t vc_id, kbp_rqt_data_t * req_p)
{
	int32_t ret_val = KBP_FAIL;

	kbp_print_debug ("Inside KBP Send\n");

	if (KBP_DIR_REQ_DIR_RSP_TYPE == req_p->req_rsp_type)
	{

		ret_val = kbp_pvt_send_direct_req_direct_rsp (vc_id, req_p);
	}
	else if (KBP_MEM_REQ_TYPE == req_p->req_rsp_type)
	{

		ret_val = kbp_pvt_send_mem_req (vc_id, req_p);
	}
	else
	{
		kbp_print_info ("NOT Supported As of Now \n");
	}

	return ret_val;
}


/**
* @brief: kbp_recv function is used to receive the response
*		  from the kbp for previous request associated with the given vc.
*
* @param[in]  vc_id     :virtual channle id
* @param[in]  recv_rsp    :pointer to kbp response data structure
*
* @return
*  - On sucess,  returns 0
*  - On failure, returns -1
*
* @sa kbp_recv
* @ingroup kbp
*
*/
int32_t
kbp_recv (int vc_id, kbp_rsp_data_t * recv_rsp)
{
	uint32_t size = 0, src = 0, code = 0;
	uint32_t rsptype = 0, i = 0;
	uint64_t rsp_data[4] = { 0x0ULL, };
	int32_t ret_val = KBP_OK;
	do
	{
		kbp_print_debug ("Message Wait\n");
		xlp_message_wait (0x1);

		i = xlp_message_receive (vc_id, &src, &size, &code, &rsp_data[0],
					 &rsp_data[1], &rsp_data[2], &rsp_data[3]);

		kbp_print_debug
				("Cpu received (cpu id = %d ret val = %d) a message of : \
				src %d, size %d, { 0x%lx 0x%lx 0x%lx 0x%lx }\n",
				vc_id, i, src, size, rsp_data[0], rsp_data[1],
				rsp_data[2], rsp_data[3]);
	}
	while (src != ILA_STATION_ID);


	rsptype = (rsp_data[0] & DESC_TYPE_MASK) >> 63;

	recv_rsp->recv_msg_size = size;
	recv_rsp->rsp_type = rsptype;

	if (rsptype == CAM_DESC_DIRECT)
	{
		recv_rsp->req_id = (rsp_data[0] & 0xFFFC000000000000) >> 50;

		recv_rsp->recv_msg[0] = (rsp_data[1] & 0xFFFFFFFF00000000)
						| (rsp_data[0] & 0x00000000FFFFFFFF);
		recv_rsp->recv_msg[1] = (rsp_data[2] & 0xFFFFFFFF00000000)
						| (rsp_data[1] & 0x00000000FFFFFFFF);

		recv_rsp->recv_msg[2] = (rsp_data[3] & 0xFFFFFFFF00000000)
						| (rsp_data[2] & 0x00000000FFFFFFFF);

		recv_rsp->req_header = ((rsp_data[0] >> 32) & 0x3FFFFFFF);

		/* Handle the error - Bit 7 and 8 denotes the error status */
		if (rsp_data[0] & 0x0000018000000000)
			ret_val = kbp_handle_error (vc_id, recv_rsp);
	}
	else if (rsptype == CAM_DESC_MEM)
	{
		recv_rsp->req_id = (rsp_data[0] >> 48) & 0xFFF;
		recv_rsp->rsp_len = (rsp_data[0] >> 40) & 0xFF;
		recv_rsp->recv_msg[0] = rsp_data[0] & 0xFFFFFFFFFF;
	}

	return ret_val;
}

/**
* @brief: kbp_destroy function is used to initialize the kbp link and other
*		 data structures required.
*
* @param[in]  vc_id     		:virtual channle id
* @param[in]  destroy_link    :Specifies that link needs to be destroy or not
*
* @return
*  - None
*
* @sa kbp_destroy
* @ingroup kbp
*
*/
void
kbp_destroy (uint32_t vc_id, uint32_t destroy_link)
{
	(void)destroy_link;
	kbp_pvt_destroy_mempool (&g_kbp_mempool);
	return;
}

