/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#include <stdio.h>

#include "types.h"
#include "bridge.h"
#include "iomap.h"
#include "uart.h"
#include "byteorder.h"
#include "pt_regs.h"
#include "fifo.h"
#include "shell.h"
#include "vuart.h"
#include "mips-exts.h"
#include "mm.h"

static struct fifo tx_fifo;
static struct fifo rx_fifo;

extern unsigned int xlr_netos_uart;

#define TX_BURST_SIZE 1

#define DMESG_BUFSIZE 0x1000
unsigned char dmesg_buf[DMESG_BUFSIZE];
int dmesg_buf_index;
int dmesg_trace = 0;

static int uart_init_done;

#define RING_BUFFER_BASE (511 << 20)
#define RING_BUFFER_SIZE (8 << 10)
static void outbyte_ring_buffer(char c)
{
	unsigned long base = RING_BUFFER_BASE + (processor_id() * RING_BUFFER_SIZE);
	char *buf = (char *)KSEG0 + base;
	static int idx = 0;

	buf[idx] = c;
	idx = (idx + 1) % (RING_BUFFER_SIZE);
}

void uart_outbyte(phoenix_reg_t *mmio, int data)
{
	if (!dmesg_trace) {

		if (!uart_init_done) outbyte_ring_buffer(data);

		phoenix_write_reg(mmio,UART_THR, cpu_to_be32(data));
	}
	else {
		if (dmesg_buf_index < DMESG_BUFSIZE)
			dmesg_buf[dmesg_buf_index++] = (unsigned char)data;
	}
}

int inbyte_nowait(void)
{
	int    data;
	int    i, lsr = 0;
	volatile uint32_t *mmio;

	if (xlr_netos_uart == VIRT_UART){
		data = vuart_inbyte_nowait();
		return data;
	}
	else if(xlr_netos_uart == TTYS1){
		mmio = phoenix_io_mmio(PHOENIX_IO_UART_1_OFFSET);
	}else{
		mmio = phoenix_io_mmio(PHOENIX_IO_UART_0_OFFSET);
	}

	data = EOF;

	if (fifo_dequeue(&rx_fifo, &data)) { /* characters to be read already in fifo */
		return data;
	}

	do {
		lsr = be32_to_cpu(phoenix_read_reg(mmio,UART_LSR));

		if (lsr & 0x20) {
			/* Tx available, try to send any characters */
			for (i = 0; i < TX_BURST_SIZE && fifo_dequeue(&tx_fifo, &data); i++) {
				outbyte(data);
			}
		}
		if (lsr & 0x80) { /* parity/frame/break - push a 0! */
			break;
		}
		if (lsr & 0x01) { /* Rx Data */
			data = be32_to_cpu(phoenix_read_reg(mmio,UART_RHR));
			break;
		}
	} while(0);

	return data;
}

#include "shared_structs.h"

void outbyte(char c)
{
	int lsr, _stack[2], _top;

	volatile uint32_t *mmio ;

	if (!uart_init_done) {
		outbyte_ring_buffer(c);
		return;
	}

	if (xlr_netos_uart == VIRT_UART){
		vuart_outbyte(c);
		return;
	}
	else if(xlr_netos_uart == TTYS1){
		mmio = phoenix_io_mmio(PHOENIX_IO_UART_1_OFFSET);
	}
	else{
		mmio = phoenix_io_mmio(PHOENIX_IO_UART_0_OFFSET);
	}

	lsr = _top = 0; 

	_stack[_top] = c;
	if (c == '\n') { 
		_top++;
		_stack[_top] = '\r'; 
	}

	for ( ; _top >= 0; _top--) {
		do {
			lsr = be32_to_cpu(phoenix_read_reg(mmio,UART_LSR));
			if (lsr & 0x80) { 
				fifo_enqueue(&rx_fifo, 0);
			}
			if (lsr & 0x01) { 
				fifo_enqueue(&rx_fifo, 
						be32_to_cpu(phoenix_read_reg(mmio,UART_RHR)));
			}
		} while (!(lsr & 0x20)); 

		phoenix_write_reg(mmio,UART_THR, cpu_to_be32(_stack[_top]));
	}
	return;
}

char inbyte(void)
{
    char ch;
    int    data = 0;
    int    i, lsr = 0;
    volatile uint32_t *mmio;

    if (xlr_netos_uart == VIRT_UART)    {
        ch = vuart_inbyte();
        return ch;
    }
    else if(xlr_netos_uart == TTYS1){
	    mmio = phoenix_io_mmio(PHOENIX_IO_UART_1_OFFSET);
    }
    else{
	    mmio = phoenix_io_mmio(PHOENIX_IO_UART_0_OFFSET);
    }

    if (fifo_dequeue(&rx_fifo, &data)) {    /* characters to be read already in fifo */
        return (char)data;
    }
    
    for ( ; ; ) {
        lsr = be32_to_cpu(phoenix_read_reg(mmio,UART_LSR));
        
        if (lsr & 0x20) {    /* Tx available, try to send any characters */
            for (i = 0; i < TX_BURST_SIZE && fifo_dequeue(&tx_fifo, &data); i++) {
                uart_outbyte(mmio, data);
            }
        }
        if (lsr & 0x80) {    /* parity/frame/break - push a 0! */
            if (!fifo_enqueue(&rx_fifo, 0)) 
                break;      
        }
        if (lsr & 0x01) {    /* Rx Data */
            data = be32_to_cpu(phoenix_read_reg(mmio,UART_RHR));
            if (!fifo_enqueue(&rx_fifo, data))
                break;      
        }
        if (!fifo_empty(&rx_fifo)) 
            break;
    }

    fifo_dequeue(&rx_fifo, &data);

    return (char)data;
}

void uart_int_handler(int irq, struct pt_regs *regs)
{
    return;
}

void uart_init(void)
{
    volatile uint32_t *mmio = phoenix_io_mmio(PHOENIX_IO_UART_0_OFFSET);

    uart_init_done = 1;

    if (xlr_netos_uart == VIRT_UART)    {
        vuart_init();
        return;
    }
    else if (xlr_netos_uart == TTYS1)	{
	    mmio = phoenix_io_mmio(PHOENIX_IO_UART_1_OFFSET);
    }
    else	{
	    mmio = phoenix_io_mmio(PHOENIX_IO_UART_0_OFFSET);
    }

    fifo_init(&rx_fifo);
    fifo_init(&tx_fifo);
}

/* ================================================================================*/

static void dmesg_shell_cmd(int argc, char *argv[])
{
    int i=0;

    for(i=0;i<dmesg_buf_index;i++)
        outbyte((char)dmesg_buf[i]);  
    outbyte('\n');
}

void dmesg_on(void)
{
#if 0
	dmesg_trace = 1;
#endif
}

void dmesg_off(void)
{
	dmesg_trace = 0;
}

void console_init(void)
{
	register_shell_cmd("dmesg", dmesg_shell_cmd, "Print Bootup messages");
	dmesg_on();
}
