/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

/**
* @file_name nae.c
*/

/**
* @defgroup nae Network Accelerator Engine (NAE) API
* @brief Description about the NAE apis supported in Hyperexec
*/

#include <reent.h>
#include <errno.h>
#include <smp.h>
#include <spinlock.h>
#include <devices.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <mm.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <nae.h>
#include <nlm_hal_nae.h>
#include <nlm_hal_fmn.h>
#include "fdt_local.h"
#include <nlm_dev.h>

static int rx_vc, fb_vc;
static spinlock_t nae_lock;
static int cur_if = 0;
static int skip_rx_buf_alloc;
#define NAE_NAME_SZ 50
#define NAE_INIT_SHM_NAME "nae-init-shm"
#define NAE_DATA_SHM_NAME "nae-mem"
static nae_port_t *nae_ports;

//#define DEV_DBG 1

static int get_iface(int fd)
{
	struct device_ops *dev;
	int iface;
	if ((dev = dev_get(fd))) {
		sscanf(dev->name, NAE_IF_FMT, &iface);
		return iface;
	}
    /* Return Invalid Interface Number */
	return MAX_SGMII_PORTS;
}
/*
 *      The Main API
 */
int nae_open(struct _reent *ptr, const char *file, 
		int flags, int mode) 
{
	int iface, hw_iface;
	int nae_rx_vc;
	unsigned long long desc;
	unsigned long addr;
	int i;

	sscanf(file, NAE_IF_FMT, &iface);
	hw_iface = nae_cfg.ports[iface].hw_port_id;

	printf("nae_if%d: Open \n", iface);

	nae_rx_vc = XLP_NET_RX_VC_BASE + hw_iface;

	if (nae_ports[iface].free_desc_mem == NULL)
		goto nae_open_exit;

	/* Send free descriptors */
	for (i = 0; i < nae_ports[iface].num_free_desc; ++i) { 


		addr = (unsigned long) mpalloc(nae_ports[iface].free_desc_mem, 0);
		if (addr == 0) {
			printf("Failed to allocate Free Desc\n");
			return -1;
		}

                /* Actual packet data start after NAE_DESC_HDR_SIZE bytes. */
		addr += NAE_DESC_HDR_SIZE; 
		desc = (mpgetphys(nae_ports[iface].free_desc_mem,(void *)addr)) & 0xffffffffc0ULL;
		if (nlm_hal_send_msg1( nae_rx_vc,
							   0,
							   desc)) {

			printf("%s; Failed to send Free-in desc\n", file);
			return -1;
		}
	}

	nlm_hal_mac_enable(nae_cfg.ports[iface].hw_port_id, nae_cfg.ports[iface].iftype);
nae_open_exit:
	/* TODO: To reserve some fd's for NAE */
	return NAE_FD_BASE + iface;
}

static int nae_close(struct _reent *ptr, int fd)
{
	printf("nae_if%d: Close \n", get_iface(fd));
	return nlm_hal_close_if(SGMII_IF, get_iface(fd));
}

static void nae_init(void)
{
	printf("nae_if%d: Init\n", cur_if);
	cur_if++;
	smp_lock_init(&nae_lock);
}

void *alloc_nae_mem(int num_ports, int start_port)
{
	int rv, mem_fd;
	int total_sz = sizeof(struct device_ops);
	void *nae_mem;

	total_sz += sizeof(nae_port_t);
	total_sz += NAE_NAME_SZ; /* To account for name str */
	total_sz *= num_ports;
 
	/* Create a shared memory object and set its size */
	mem_fd = shm_open(NAE_DATA_SHM_NAME, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	if(mem_fd < 0) {
		printf("NAE_CFG_ERROR: Failed to allocate NAE shared mem\n");
		return NULL;
	}
 
	/* Set the size   */
	rv = ftruncate(mem_fd, total_sz);
	if(rv < 0) {
		printf("NAE_CFG_ERROR: Failed to set size of NAE shared mem\n");
		return NULL;
	}
 
	/* Map the shared memory object */
	nae_mem = (unsigned char *) mmap(NULL, total_sz,
				       PROT_READ | PROT_WRITE,
				       MAP_SHARED, mem_fd, 0);
	if((void *) nae_mem == MAP_FAILED) {
		printf("NAE_CFG_ERROR: Failed to mmap NAE shared mem\n");
		return NULL;
	}

	/* Carve out the different structures. */
	nae_ports = (nae_port_t *)(nae_mem + (num_ports * sizeof(struct device_ops)) + (num_ports * NAE_NAME_SZ));
	return (void *)nae_mem;
}



static int alloc_nae_ports(int num_ports, int start_port)
{
	int i, fd;
	char *name;
	struct device_ops *dev;

	if((dev = alloc_nae_mem(num_ports, start_port)) == NULL)
		return -1;
	

	name = (void *)dev + (num_ports * sizeof(struct device_ops));
	memset(nae_ports, 0, (num_ports * sizeof(nae_port_t)));
	cur_if = start_port;
	for (i = 0 ; i < num_ports; ++i) {
		sprintf(name, NAE_IF_FMT, (i + start_port));
		dev[i].name = name;
		dev[i].init = nae_init;
		dev[i].open_r = nae_open;
		dev[i].close_r = nae_close;
		dev[i].obj = &dev[i]; /* Temp hack */

		if(dev[i].obj == NULL) 
			return -1;

		fd = NAE_FD_BASE + i; /* NAE FDS start from NAE_FD_BASE */
		if(dev_add((struct device_ops *)&dev[i], fd) < 0) {
			fdset(fd);
			return -1;
		}
		name += NAE_NAME_SZ;

		nae_ports[i].id = start_port + i;
	}

	return 0;
}
/* To allocate freein desc memory
 * TODO: currently only supports uniqe desc size for all interfaces
 *       Need to support differnet sizes. For that need to replace mpool API
 * with shm_xyz API
 */
static int alloc_nae_free_desc_mem(int intf, int size)
{
	int esize = size + NAE_DESC_HDR_SIZE;
	mpoolptr_t mem;

	if (nae_ports[intf].init_done) {
		return 0;
	}

	mem = mpcreate("nae_free_desc_mem", nae_ports[intf].num_free_desc, esize, XLP_CACHE_LINE_SIZE, 0);
	if (mem == NULL) {
		printf("NAE Init: Failed to allocate Free Desc Memory for interface %d\n", intf);
		return -1;
	}
	nae_ports[intf].free_desc_mem = mem;
	nae_ports[intf].init_done = 1;
	return 0;
}
/* Temporarily placing the soc_recv here. Should move it to a common file
 * Currently Only supports NAE !!
 */
int soc_recv(int vc, unsigned int *intf, uint64_t *paddr, unsigned int *flags)
{
	uint32_t size, code = 0;
        unsigned rx_status = 0;
        int cpu = getpid(); 
        int dst_vc = 4 * cpu + vc;
        uint64_t data[4];
        int len = 0;

        if(nlm_hal_recv_msg2( dst_vc, 
                              (uint32_t *) intf,
                              &size, 
                              &code,
                              &data[0],
                              &data[1])) {

                rx_status = xlp_read_rx_status();
                printf("[%d] rx failed for vc %d\n", cpu, vc);

                if (!((rx_status >> 28) & (1 << dst_vc))) {
                        printf("Receive failed1 (rx_status=%08x)\n", rx_status);

                }

                printf("Receive failed2 (rx_status=%08x)\n", rx_status);
                return len;
        }

        *paddr = (data[1]) & 0xffffffffc0ULL;
        len = (data[1] >> 40) & 0x3fff;

#ifdef DEV_DBG
        printf("%s.%d data %lx %lx\n", __func__, cpu, data[0], data[1]);
        printf("%s.%d Addr %lx len %d\n", __func__, cpu, *paddr, len);
#endif

        return len;
}

int _nae_recv_poe(int vc, uint64_t *poe_hdr, unsigned int *intf, uint64_t *paddr, int *cntl)
{
	uint32_t size = 0, code = 0;
	unsigned int rx_status = 0;
	int dst_vc = vc;
	uint64_t data[4];
	int len = 0;

	if(nlm_hal_recv_msg2( dst_vc, 
			      (uint32_t*) intf,
			      &size, 
			      &code, 
			      &data[0], 
			      &data[1])) {

		rx_status = xlp_read_rx_status();
		/* printf("[%d] rx failed for vc %d\n", cpu, vc);    */

		if (!((rx_status >> 28) & (1 << dst_vc))) {
			/* printf("Receive failed1 (rx_status=%08x)\n", rx_status); */

		}

		/* printf("Receive failed2 (rx_status=%08x)\n", rx_status); */
		return 0;
	}

	*poe_hdr = data[0];

	if (size == 1) { /* Free Descriptor */
		unsigned int err = (data[0] >> 54) & 0x3;
		printf("FD: 0x%llx size %d\n", (unsigned long long) data[0], size);
		if (err) {
			return -NLM_NAE_RX_TXERR;
		}
		return -NLM_NAE_RX_TXC;
	}

	*paddr = (data[1]) & 0xffffffffc0ULL;
	/* Ignore any stray packets received */
	if ((*paddr  >= (255 << 20)) && (*paddr < (256 << 20))) {
		return 0;
	}

	*cntl = (data[1] >> 54) & 0x3ff; /* Hack ! Overloading flags to send context id to app */

        if (data[1] & 0x10) { /* RXError Detected */
		/* printf("RE: 0x%llx\n", (unsigned long long) data[1]); */
		return -NLM_NAE_RX_ERR;
	}

	len = (data[1] >> 40) & 0x3fff;

	/* Ignore an unexpected p2p pkt */
	if (data[1] & 0x1) { 
		/* printf("Addr of p2p pkt 0x%llx\n, (unsigned long long) *paddr); */
		return 0;
	}

#ifdef DEV_DBG
	printf("%s size %d context %d \n", __func__,  size, *cntl);
	printf("%s.%d data %lx %lx\n", __func__, cpu, data[0], data[1]);
	printf("%s.%d Addr %lx len %d\n", __func__, cpu, *paddr, len);
#endif
	len -= 4; /* To account for ethernet CRC */

	if ((len < 0) || (len >1518)) {
		printf("LE: 0x%llx\n", (unsigned long long) data[1]);
		return -NLM_NAE_RX_ERR;
	}

	return len;
}

int _nae_recv(int vc, unsigned int *intf, uint64_t *paddr, unsigned int *cntl)
{
	uint32_t size = 0, code = 0;
	unsigned int rx_status = 0;
	int dst_vc = vc;
	uint64_t data[4];
	int len = 0;

	if(nlm_hal_recv_msg2( dst_vc, 
			      (uint32_t*) intf,
			      &size, 
			      &code, 
			      &data[0], 
			      &data[1])) {

		rx_status = xlp_read_rx_status();
		/* printf("[%d] rx failed for vc %d\n", cpu, vc);    */

		if (!((rx_status >> 28) & (1 << dst_vc))) {
			/* printf("Receive failed1 (rx_status=%08x)\n", rx_status); */

		}

		/* printf("Receive failed2 (rx_status=%08x)\n", rx_status); */
		return 0;
	}

	if (size == 1) { /* Free Descriptor */
		unsigned int err = (data[0] >> 54) & 0x3;
		/* printf("FD: 0x%llx\n", (unsigned long long) data[0]); */
		if (err) {
			return -NLM_NAE_RX_TXERR;
		}
		return -NLM_NAE_RX_TXC;
	}

	*paddr = (data[1]) & 0xffffffffc0ULL;
	/* Ignore any stray packets received */
	if ((*paddr  >= (255 << 20)) && (*paddr < (256 << 20))) {
		return 0;
	}

	*cntl = (data[1] >> 54) & 0x3ff; /* Hack ! Overloading flags to send context id to app */

        if (data[1] & 0x10) { /* RXError Detected */
		/* printf("RE: 0x%llx\n", (unsigned long long) data[1]); */
		return -NLM_NAE_RX_ERR;
	}

	len = (data[1] >> 40) & 0x3fff;

	/* Ignore an unexpected p2p pkt */
	if (data[1] & 0x1) { 
		/* printf("Addr of p2p pkt 0x%llx\n, (unsigned long long) *paddr); */
		return 0;
	}

#ifdef DEV_DBG
	printf("%s.%d size %d context %d \n", __func__, cpu, size, *cntl);
	printf("%s.%d data %lx %lx\n", __func__, cpu, data[0], data[1]);
	printf("%s.%d Addr %lx len %d\n", __func__, cpu, *paddr, len);
#endif
	len -= 4; /* To account for ethernet CRC */

	if ((len < 0) || (len >1518)) {
		printf("LE: 0x%llx\n", (unsigned long long) data[1]);
		return -NLM_NAE_RX_ERR;
	}

	return len;
}
/**
* @brief nae_recv function is used to receive data at a physical address location.
*
* @param[in]  paddr    :Physical address where the data is received
* @param[out]  cntl    :context number of the received packet
*
* @return
*  - On sucess,  returns >0 (length of the packet)
*  - On failure, returns 0 (no message received)
*  - On failure, returns <0 (reason for failure). Following errors are detected
*             -   -NLM_NAE_RX_TXERR : Unexpected TX complete with Errors message
*             -   -NLM_NAE_RX_TXC   : Unexpected TX Complete message
*             -   -NLM_NAE_RX_ERR   : Packet received with RX Error
* 
* @sa nae_send
* @ingroup nae
*
*/

int nae_recv(uint64_t *paddr, unsigned int *cntl)
{
	unsigned int intf;
	return _nae_recv(rx_vc, &intf, paddr, cntl);
}

int nae_recv_poe(uint64_t *poe_hdr, uint64_t *paddr, int *cntl)
{
	unsigned int intf;
	return _nae_recv_poe(rx_vc, poe_hdr, &intf, paddr, cntl);
}
/* nae_send
 */
static __inline__ uint64_t nae_tx_desc(unsigned int type,
				       unsigned int rdex,
				       unsigned int fbid,
				       unsigned int len,
				       uint64_t addr)
{
	return ((uint64_t)(type & 0x3) << 62) |
		((uint64_t)(rdex & 0x1) << 61) |
		((uint64_t)(fbid & 0x7f) << 54) |
		((uint64_t)(len & 0x3fff) << 40) |
		addr;
}
/**
* @brief nae_send function is used to send "len" bytes of data located at specified physical address
*         to dest.
*
* @param[in]  dest      :Destination context number on which packet needs to be transmitted.
* @param[in]  paddr     :Physical address
* @param[in]  len       :Length of data 
* @param[in]  flags     :The following flags are supported
*                 -      NLM_NAE_TX_FREE_IN  : Send a free descriptor 
* @return
*  - On sucess,  returns 0.
*  - On failure, returns -1.
*
* @note
*    nae_recv will return the context number which can be used directly by the nae_send
*
* @sa nae_recv
* @ingroup nae
*
*/

int nae_send(int dest, int fbid, uint64_t paddr, int len, unsigned int flags)
{
	uint64_t data[4];

	if (flags & NLM_NAE_TX_FREE_IN) {

		unsigned int nae_rx_vc = XLP_NET_RX_VC_BASE + dest;

		xlp_message_send_block_fast_1(0 /* code */,
					      nae_rx_vc,
					      (paddr & 0xffffffffc0ULL));

		return 0;
	}

#ifdef DEV_DBG
	printf("%s dest %d ldaddr %llx\n",
	       __func__, dest, (unsigned long long) paddr);
#endif

	data[0] = nae_tx_desc(P2D_NEOP, 0, fbid, 0, paddr);
	data[1] = nae_tx_desc(P2D_EOP, 
			      0, 
			      NULL_VFBID, 
			      len , 
			      paddr );

	/* printf("%s.%d sending to %d fbid %d \n", __func__, getpid(), dest + XLP_NET_TX_VC_BASE, fbid); */
	xlp_message_send_block_fast_2(0, /* code */
				      dest + XLP_NET_TX_VC_BASE,
				      data[0], 
				      data[1]);

	return 0;
}

int nae_send_poe(int fwd_mode, uint64_t poe_hdr, int poe_class, int fbid, uint64_t paddr, int len)
{
	uint64_t data[3];
	int poe_dst_vc = XLP_POE_VC_BASE + poe_class;

#ifdef DEV_DBG
	printf("mode %d hdr %llx, cl %d, fbid %d paddr %llx len %d\n", fwd_mode, poe_hdr, poe_class, fbid, paddr, len);
	printf("%s dest %d ldaddr %llx\n",
	       __func__, dest, (unsigned long long) paddr);
#endif

	data[0] = poe_hdr;
	data[1] = nae_tx_desc(P2D_EOP, 
			      0, 
			      fbid, 
			      len , 
			      paddr );

	xlp_message_send_block_fast_2(fwd_mode, /* code */
				      poe_dst_vc,
				      data[0], 
				      data[1]);

	return 0;
}

static inline void *nlm_nae_shm_init(int *shmfd)
{
	int fd, rv;
	struct nae_shm_vars *m;
	const char *name = NAE_INIT_SHM_NAME;

	fd = shm_open(name, O_CREAT | O_RDWR, 0600);
	if(fd < 0) {
		printf("shm_open failed for %s\n", name);
		return NULL;
	}

	rv = ftruncate(fd, sizeof(struct nae_shm_vars));
	if(rv < 0) {
		printf("ftruncate failed on fd %d for %s\n", fd, name);
		goto err_exit;
	}

	m = mmap(NULL, sizeof(struct nae_shm_vars), PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if(m == MAP_FAILED) {
		printf("mmap failed on fd %d for %s\n", fd, name);
		goto err_exit;
	}

	*shmfd = fd;
	return m;

err_exit:
	shm_unlink(name);
	return NULL;
}

extern int dom_id;

/* The following macro is used by init_nae. It may return to the caller.
 * Necessary bookkeeping is needed.
 */
#define NAE_GET_FPROP(fdt, path, prop, prop_buf)					\
	if (get_fdt_prop(fdt, path, prop, prop_buf, PROP_INT32) != 0) {			\
		fprintf(stderr, "Get prop (%s) failed on %s, Aborting\n", prop, path);	\
		spin_unlock(&m->init_lock);						\
		return -1;								\
	}

int init_nae(void *fdt)
{
	int num_nae_ports, start_port;
	char if_node_str[50];
	char cpu_node_str[32];
	int i, rv = 0, shmfd;
	struct nae_shm_vars *m;
	int free_desc_size;
	extern int nlm_nae_owner;

	if ((m = nlm_nae_shm_init(&shmfd)) == NULL) {
		printf("Failed to initialze nae shared memory");
		return -1;
	}

	spin_lock(&m->init_lock);
	if (m->init_done) {
		spin_unlock(&m->init_lock);
		munmap(m, sizeof(struct nae_shm_vars));
		close(shmfd);
		return 0;
	} else {
		m->init_done = 1;
	}

        /* Configure credits to non-n0c0 cores */
        nlm_hal_fmn_init(0x10000000, 0x02000000, 50);

	if(nlm_nae_owner)
		nlm_hal_init_nae(fdt, dom_id);

	/* 
	 *     Parse the rest of Nae Config from FDT
	 */
	snprintf(cpu_node_str, 32, "/doms/dom@%d/cpu", dom_id);
	NAE_GET_FPROP(fdt, cpu_node_str, "nae-rx-vc", &rx_vc);
	NAE_GET_FPROP(fdt, cpu_node_str, "nae-fb-vc", &fb_vc);
	NAE_GET_FPROP(fdt, cpu_node_str, "nae-buf-alloc-by-app", &skip_rx_buf_alloc);

	NAE_GET_FPROP(fdt, NAE_CFG_NODE, "num-ports", &num_nae_ports);
	NAE_GET_FPROP(fdt, NAE_CFG_NODE, "start-port-id", &start_port);

	if(!nlm_nae_owner){
		printf("Skipping NAE initalization.\n");
		goto nae_err_exit;
	}

	if (alloc_nae_ports(num_nae_ports, start_port) < 0) {
		printf("Allocation of NAE interfaces Failed \n");
		rv = -1;
		goto nae_err_exit;
	}
	for (i = 0; i < num_nae_ports; ++i) {
		sprintf(if_node_str, "%s/port@%d", NAE_CFG_NODE, (start_port + i));

		NAE_GET_FPROP(fdt, if_node_str, "num-free-desc", &nae_ports[i].num_free_desc);
		NAE_GET_FPROP(fdt, if_node_str, "free-desc-size", &free_desc_size);

		if (skip_rx_buf_alloc) {
			nae_ports[i].free_desc_mem = NULL;
			printf("Nae Config: Skipping Pkt Buf allocation for port %d\n", i);
			continue;
		}

		/* Alloc NAE Free Desc */
		if (alloc_nae_free_desc_mem(i, free_desc_size)) { 
			rv = -1;
			goto nae_err_exit;
		}
	}

nae_err_exit:
	spin_unlock(&m->init_lock);
	return rv;
}

void nlm_dealloc_shmem(void)
{
	shm_unlink(NAE_DATA_SHM_NAME);
	shm_unlink(NAE_INIT_SHM_NAME);
}

int nae_set_poe_distvec(int vec, uint32_t cm0, uint32_t cm1, uint32_t cm2, uint32_t cm3, uint32_t vcmask)
{
	return nlm_hal_init_poe_distvec(vec, cm0, cm1, cm2, cm3, vcmask);
}

uint16_t nae_get_hwport(uint32_t context)
{
	return nlm_hal_get_hwport(context);
}
