/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

// XLR network driver common
// device based interface to 
// gmac, xgmac & spi4 devices

// Using this interface results in 5 %
// lower performance than using the 
// functions directly from lower layers

// includes
#include <stdio.h>
#include <stdlib.h>

#include "pci.h"
#include "types.h"
#include "mips-exts.h"
#include "asmmacros.h"
#include "mipsregs.h"
#include "iomap.h"
#include "netdevice.h"
#include "spinlock.h"
#include "packet.h"
#include "pod.h"
#include "mac.h"
#include "xgmac.h"
#include "spi4.h"
#include "net_config.h"
#include "msgring.h"
#include "mm.h"

static struct driver_data    macs[PHOENIX_MAX_MACS];

// internal use function to return tx bucket, given device and port number
static int msgrng_tx_bucket (int id, int type, int port) {
	int txid;
	if      (type == GMAC)    txid = msgrng_gmac_stid_tx(id);
	else if (type == XGMAC)   txid = msgrng_xgmac_stid_tx(id);
	else if (type == SPI4)    txid = msgrng_xgmac_stid_tx(id) + port;
	else                      txid = 127;
	return txid;
}

// internal use function to return free back to mac bucket
static int msgrng_free_bucket (int id, int type, int port) {
	int txid;
	if      (type == GMAC)    txid = msgrng_gmac_stid_rfr(id);
	else if (type == XGMAC)   txid = msgrng_xgmac_stid_rfr(id);
	else if (type == SPI4)    txid = msgrng_xgmac_stid_rfr(id);
	else                      txid = 127;
	return txid;
}

// access functions
int mac_tx (struct packet *pbuf, struct net_device *dev, int len, int port) {
	struct driver_data *data = dev->driver_data;
	int txid = 0;
	uint8_t *pkt_buf = &pbuf->data[0];
	
	if ( !pkt_valid(pbuf) ) {
		printf("[%s]: Transmitting bad packet Address! dropping it\n", __FUNCTION__);    
		return -1;
	}
	
	struct msgrng_msg msg;
	mac_make_desc_tx(&msg, data->id, data->type, virt_to_phys(pkt_buf), len);  
	txid = msgrng_tx_bucket(data->id, data->type, port);
	if (message_send_1(MSGRNG_CODE_MAC, txid, &msg) != 0) {
		return -1;
	}  
	return 1;
}


// transmit w/ free back to mac's free-in pool
inline int mac_tx_free (struct packet *pbuf, struct net_device *dev, int len, int port) {
	struct driver_data *data = dev->driver_data;
	int txid;
	int fbid;
	uint8_t *pkt_buf = &pbuf->data[0];
	unsigned long addr = virt_to_phys(pkt_buf);
	struct msgrng_msg msg;
	txid = msgrng_tx_bucket( data->id, data->type, port);
	fbid = msgrng_free_bucket( data->id, data->type, port);
	
	msg.msg0 =      ( ((uint64_t)1  << 63) |
				            ((uint64_t)fbid << 54) |
				            ((uint64_t)len << 40)  |
				            ((uint64_t)addr));
 
	message_send_block_fast_1( MSGRNG_CODE_MAC, txid, msg.msg0);
	
	/* Let the mac keep the free descriptor */
	return 1;
}



static int mac_rx (void (*netif_rx)(struct packet *pbuf), struct net_device *dev, int bucket) {
	int    size = 0;
	int    code = 0;
	int    stid = 0;
	int    length =0;
	struct msgrng_msg    msg;
	struct packet       *pbuf = 0;
	int                  ctrl = 0;
	//int                  bucket = (processor_id() & 0x03);
	
	// wait indefinitly for a message
	for (;;) {
		msgrng_wait((1 << bucket));
		if (message_receive(bucket, &size, &code, &stid, &msg)) 
			continue;    
		pbuf = (struct packet *)((char *)phys_to_virt((msg.msg0 & 0xffffffffe0ULL)) 
				   - sizeof(unsigned short));
		
		ctrl   = (msg.msg0 >> 63) & 0x01;
		length = (msg.msg0 >> 40) & 0x3fff;
		if (length == 0) {
			free_pbuf(pbuf);
			continue;
		};
		pbuf->len = (uint16_t)(msg.msg0 >> 40) & 0x3fff;
		
		//printf("[%s]: Rx Msg of size=%d, addr=%p, len=%d from stid = %d\n", 
		//   __FUNCTION__, size, pbuf->data, pbuf->len, stid);
		if (!pkt_valid(pbuf)) {
			printf("Bad Packet Address received! dropping it\n");
			continue;
		}
		
		// pass packet buffer to upper layer
		netif_rx (pbuf);
	}
	return 0;
}



// rx wait function
struct packet *mac_rx_wait(uint32_t *cycles, struct net_device *dev, 
				 unsigned int *len, unsigned int *seq_num, int *port, int bucket){
	int                    size, code, stid;
	struct msgrng_msg      msg;
	struct packet         *pbuf = 0;
	int                    ctrl = 0;
	int                    length =0;
	uint32_t               counter = 0;
	struct driver_data    *data = dev->driver_data;
	int                    id;
	
	size = code = stid = counter = 0;
	id = processor_id();
	//bucket = id & 0x03;
	
	/* Wait indefinitely for a msg */
	for ( ; ; ) {
		msgrng_wait(1 << bucket);
		counter = read_32bit_cp0_register(CP0_COUNT);
		if (message_receive(bucket, &size, &code, &stid, &msg)) 
			continue;
	
		pbuf = (struct packet *) ((char *) phys_to_virt( (msg.msg0 & 0xffffffffe0ULL)) );
		ctrl = (msg.msg0 >> 61)  &  0x07;
		length = (msg.msg0 >> 40) & 0x3fff;
		if (length == 0) {
			stid = mac_make_desc_rfr(&msg, data->id, data->type, virt_to_phys(pbuf->data));
			//printf("cpu_%d: Sending (ctrl=%d) rfr @ (%lx) to station id %d\n", 
			//     id, ctrl, virt_to_phys(pbuf->data), stid);
			message_send_block(1, MSGRNG_CODE_MAC, stid, &msg);
			continue;
		}
		*seq_num = ( (msg.msg0 & 0x1f) << 12) | ((msg.msg0 >> 40) & 0xfff);
		*len = ((uint16_t) (msg.msg0 >> 40)) & 0x3fff;
		*port = (int) (msg.msg0 & 0xf);
		if ( !pkt_valid(pbuf) ) {
			printf("Bad Packet Address received! dropping it\n");
			continue;
		}
		*cycles += read_32bit_cp0_register(CP0_COUNT) - counter;
		return pbuf;
	}
	return 0;
}


// called on receivs
struct packet *mac_rx_nowait(uint32_t *cycles, struct net_device *dev, 
				   unsigned int *len, unsigned int *seq_num, 
				   int *port, int bucket){
	int size = 0;
	int code = 0;
	int stid = 0;
	struct msgrng_msg msg;
	struct packet *pbuf = 0;
	int ctrl = 0;
	int length =0;
	uint32_t counter = 0;
	int id = processor_id();
	//int bucket = id & 0x03;
	
	/* Wait indefinitely for a msg */
	for ( ; ; ) {
		counter = read_32bit_cp0_register (CP0_COUNT);
		if (message_receive(bucket, &size, &code, &stid, &msg)) 
			break;
		
		if (code == 100) {
			uint32_t *ptr = (uint32_t *)(unsigned long)msg.msg0;
			printf("cpu_%d: Received Message! size=%d code=%d bucket=%d stid=%d\n", id, size, code, bucket, stid);
			printf("cpu_%d: ptr[0] = %d\n", id, ptr[0]);
			ptr[0]++;
			printf("cpu_%d: ptr[0] = %d\n", id, ptr[0]);
			message_send_block(1, 100, 0, &msg);
			printf("cpu_%d: message sent back!\n", id);
			break;
		}
		
		/* Msg received */
		pbuf = (struct packet *) ((char *) phys_to_virt ((msg.msg0 & 0xffffffffe0ULL)) );
		ctrl = (msg.msg0 >> 63) & 0x01;
		length = (msg.msg0 >> 40) & 0x3fff;
		if (length == 0) {
			free_pbuf (pbuf);
			break;
		}
		pbuf->seq_num = *seq_num = ((msg.msg0 & 0x1f) << 12) | ((msg.msg0 >> 40) & 0xfff);
		pbuf->len = *len = ((uint16_t) (msg.msg0 >> 40)) & 0x3fff;
		*port =  (int) (msg.msg0 & 0xf);
		if (!pkt_valid(pbuf)) {
			printf("[%s]:Bad Packet Address received! dropping it, pbuf=%p\n", __FUNCTION__, pbuf);
			break;
		}
		*cycles += read_32bit_cp0_register(CP0_COUNT) - counter;
		return pbuf;
	}
	return 0;
}


// call back function ???
struct packet *mac_rx_callback(struct net_device *dev, struct msgrng_msg *msg,
				   unsigned int *len, unsigned int *seq_num){
	int stid=0;
	struct packet *pbuf = 0;
	int ctrl = 0;
	int length =0;
	struct driver_data *data = dev->driver_data;
	int id = processor_id();
	
	pbuf = (struct packet *)((char *)phys_to_virt((msg->msg0 & 0xffffffffe0ULL)));
	ctrl = (msg->msg0>>63) & 0x01;
	length = (msg->msg0>>40) &0x3fff;
	if (length ==0) {
		stid = mac_make_desc_rfr(msg, data->id, data->type, virt_to_phys(pbuf->data));
		printf("cpu_%d: Sending (ctrl=%d) rfr @ (0x%llx) to station id %d\n", 
		 id, ctrl, (unsigned long long)virt_to_phys(pbuf->data), stid);
		message_send_block(1, MSGRNG_CODE_MAC, stid, msg);
		return 0;
	}
	pbuf->seq_num = ((msg->msg0&0x1f)<<12)|((msg->msg0>>40) & 0xfff);
	pbuf->len = ((uint16_t)(msg->msg0 >> 40)) & 0x3fff;
	*seq_num = ((msg->msg0&0x1f)<<12)|((msg->msg0>>40) & 0xfff);
	*len = ((uint16_t)(msg->msg0 >> 40)) & 0x3fff;
	//printf("cpu_%d: (%s), Rx Msg addr=%p, len=%d, seq=%d \n", 
	//     id, dev->name, pbuf->data, *len, *seq_num);
	if (!pkt_valid(pbuf)) {
		printf("Bad Packet Address received! dropping it\n");
		return 0;
	}
	
	return pbuf;
}

// dev structure populate function
int xlr_dev_init(net_config *net_conf) {
	struct net_device     *dev  =0;
	struct driver_data    *priv;
	int i;
	int                    mac_devices[PHOENIX_MAX_MACS];

	mac_devices [GMAC_0]  = net_conf->gmac0 ;
	mac_devices [GMAC_1]  = net_conf->gmac1 ;
	mac_devices [GMAC_2]  = net_conf->gmac2 ;
	mac_devices [GMAC_3]  = net_conf->gmac3 ;
	mac_devices [GMAC_4]  = net_conf->gmac4 ;
	mac_devices [GMAC_5]  = net_conf->gmac5 ;
	mac_devices [GMAC_6]  = net_conf->gmac6 ;
	mac_devices [GMAC_7]  = net_conf->gmac7 ;
	mac_devices [XGMAC_0] = net_conf->xgmac0 ;
	mac_devices [XGMAC_1] = net_conf->xgmac1 ;
	mac_devices [SPI4_0]  = net_conf->spi0 ;
	mac_devices [SPI4_1]  = net_conf->spi1 ;

	for (i = 0; i < PHOENIX_MAX_MACS; i++) {
		if (mac_devices[i]) {
			dev = dev_get_by_name (phnx_mac_devices[i].name);
	 
			// driver data for each device
			priv                    = &macs[i];
			priv->dev               = dev;
			priv->mmio              = phoenix_io_mmio(phnx_mac_devices[i].io_offset);
			priv->type              = phnx_mac_devices[i].type;
			priv->iface             = phnx_mac_devices[i].instance;
			if (priv->type == XGMAC) 
				priv->id = i - XGMAC_0;
			else if (priv->type == SPI4)
				priv->id = i - SPI4_0;
			else
				priv->id = i - GMAC_0;
			
			// management function pointers
			dev->driver_data        = priv;
			dev->open               = NULL;
			dev->close              = NULL;
			dev->read_reg           = NULL;
			dev->write_reg          = NULL;
			dev->soft_reset         = NULL;
			dev->init_msgring       = NULL;
			dev->change_mac_hwaddr  = NULL;
			
			// packet datapath receive and transmit 
			dev->receive            = mac_rx;
			dev->receive_wait       = mac_rx_wait;
			dev->receive_nowait     = mac_rx_nowait;  
			dev->receive_callback   = mac_rx_callback;
			dev->transmit           = mac_tx;
			dev->transmit_free      = mac_tx_free;
		}
	}

	return 0;
}
