/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _SPINLOCK_H
#define _SPINLOCK_H

#ifndef __ASSEMBLY__

#include "mipsregs.h"
#include "mips-exts.h"
#include "types.h"

#define barrier() __asm__ __volatile__("sync": : :"memory")

typedef struct {
	volatile unsigned int lock;
} spinlock_t;

#define SPIN_LOCK_UNLOCKED (spinlock_t) { 0 }

#define spin_lock_init(x) do { (x)->lock = 0; barrier();} while(0)

#define spin_is_locked(x)       ((x)->lock != 0)
#define spin_unlock_wait(x)     do { barrier(); } while ((x)->lock)

static __inline__ void spin_lock(spinlock_t *lock)
{
	unsigned int tmp;

	__asm__ __volatile__(
	    ".set\tnoreorder\t\t\t# spin_lock\n"
	    "1:\tll\t%1, %2\n\t"
	    "bnez\t%1, 1b\n\t"
	    " li\t%1, 1\n\t"
	    "sc\t%1, %0\n\t"
	    "beqz\t%1, 1b\n\t"
	    " sync\n\t"
	    ".set\treorder"
	    : "=m" (lock->lock), "=&r" (tmp)
	    : "m" (lock->lock)
	    : "memory");
}

static __inline__ void spin_unlock(spinlock_t *lock)
{
	__asm__ __volatile__(
	    ".set\tnoreorder\t\t\t# spin_unlock\n\t"
	    "sync\n\t"
	    "sw\t$0, %0\n\t"
	    ".set\treorder"
	    : "=m" (lock->lock)
	    : "m" (lock->lock)
	    : "memory");
}

static inline unsigned int spin_trylock(spinlock_t *lock)
{
	unsigned int temp, res;

	__asm__ __volatile__(
		".set\tnoreorder\t\t\t# spin_trylock\n\t"
	    "1:\tll\t%0, %3\n\t"
	    "ori\t%2, %0, 1\n\t"
	    "sc\t%2, %1\n\t"
	    "beqz\t%2, 1b\n\t"
	    " andi\t%2, %0, 1\n\t"
	    ".set\treorder"
	    : "=&r" (temp), "=m" (lock->lock), "=&r" (res)
	    : "m" (lock->lock)
	    : "memory");

	    return res == 0;
}

/*
 * Read-write spinlocks, allowing multiple readers but only one writer.
 *
 * NOTE! it is quite common to have readers in interrupts but no interrupt
 * writers. For those circumstances we can "mix" irq-safe locks - any writer
 * needs to get a irq-safe write-lock, but readers can get non-irqsafe
 * read-locks.
 */

typedef struct {
	volatile unsigned int lock;
} rwlock_t;

#define RW_LOCK_UNLOCKED (rwlock_t) { 0 }

#define rwlock_init(x)  do { *(x) = RW_LOCK_UNLOCKED; } while(0)

#define rwlock_is_locked(x) ((x)->lock)

static inline void read_lock(rwlock_t *rw)
{
	unsigned int tmp;

	__asm__ __volatile__(
		".set\tnoreorder\t\t\t# read_lock\n"
	    "1:\tll\t%1, %2\n\t"
	    "bltz\t%1, 1b\n\t"
	    " addu\t%1, 1\n\t"
	    "sc\t%1, %0\n\t"
	    "beqz\t%1, 1b\n\t"
	    " sync\n\t"
	    ".set\treorder"
	    : "=m" (rw->lock), "=&r" (tmp)
	    : "m" (rw->lock)
	    : "memory");
}

/*  Note the use of sub, not subu which will make the kernel die with an
	overflow exception if we ever try to unlock an rwlock that is already
	unlocked or is being held by a writer.
*/

static inline void read_unlock(rwlock_t *rw)
{
	      unsigned int tmp;

	      __asm__ __volatile__(
	      ".set\tnoreorder\t\t\t# read_unlock\n"
	      "1:\tll\t%1, %2\n\t"
	      "sub\t%1, 1\n\t"
	      "sc\t%1, %0\n\t"
	      "beqz\t%1, 1b\n\t"
	      " sync\n\t"
	      ".set\treorder"
	      : "=m" (rw->lock), "=&r" (tmp)
	      : "m" (rw->lock)
	      : "memory");
}

static inline void write_lock(rwlock_t *rw)
{
	      unsigned int tmp;

	      __asm__ __volatile__(
	      ".set\tnoreorder\t\t\t# write_lock\n"
	      "1:\tll\t%1, %2\n\t"
	      "bnez\t%1, 1b\n\t"
	      " lui\t%1, 0x8000\n\t"
	      "sc\t%1, %0\n\t"
	      "beqz\t%1, 1b\n\t"
	      " sync\n\t"
	      ".set\treorder"
	      : "=m" (rw->lock), "=&r" (tmp)
	      : "m" (rw->lock)
	      : "memory");
}

static inline void write_unlock(rwlock_t *rw)
{
	      __asm__ __volatile__(
	      ".set\tnoreorder\t\t\t# write_unlock\n\t"
	      "sync\n\t"
	      "sw\t$0, %0\n\t"
	      ".set\treorder"
	      : "=m" (rw->lock)
	      : "m" (rw->lock)
	      : "memory");
}

static inline int write_trylock(rwlock_t *rw)
{
	      unsigned int tmp;
	      int ret;

	      __asm__ __volatile__(
	      ".set\tnoreorder\t\t\t# write_trylock\n"
	      "li\t%2, 0\n\t"
	      "1:\tll\t%1, %3\n\t"
	      "bnez\t%1, 2f\n\t"
	      "lui\t%1, 0x8000\n\t"
	      "sc\t%1, %0\n\t"
	      "beqz\t%1, 1b\n\t"
	      "sync\n\t"
	      "li\t%2, 1\n\t"
	      ".set\treorder\n"
	      "2:"
	      : "=m" (rw->lock), "=&r" (tmp), "=&r" (ret)
	      : "m" (rw->lock)
	      : "memory");

	      return ret;
}

static inline int read_trylock(rwlock_t *lock)
{
	      unsigned int tmp;

	      __asm__ __volatile__(
	      ".set\tnoreorder\t\t\t# read_trylock\n"
	      "1:ll\t%1, %2\n\t"
	      "bltz\t%1, 1f\n\t"
	      " addu\t%1, 1\n\t"
	      "sc\t%1, %0\n\t"
	      "beqz\t%1, 1b\n\t"
	      " sync\n\t"
	      " li %1, 0\n\t"
	      "1:nop\n\t"
	      ".set\treorder"
	      : "=m" (lock->lock), "=&r" (tmp)
	      : "m" (lock->lock)
	      : "memory");

	      return tmp==0;
}

#define NCPU 32

/* Adapted from the book 'Parallel Computer Architecture' 
 *                          - David Culler, Jaswinder Pal Singh 
 */

typedef struct {
	int counter;
	spinlock_t lock;
	uint64_t pad0[3];
	volatile int flag;
	uint32_t pad1;
	uint64_t pad2[3];
	uint64_t local_sense[NCPU];
} barrier_lock_t;

static __inline__ void barrier_lock_init(barrier_lock_t *bar)
{
	int    i;

	spin_lock_init(&bar->lock);
	bar->counter = bar->flag = 0;
	for (i = 0; i < NCPU; i++)
		bar->local_sense[i] = 0;
}

static __inline__ void barrier_lock(barrier_lock_t *bar, int count_wait)
{
	int id = processor_id();

	/* toggle cpu specific local sense */
	bar->local_sense[id] = !bar->local_sense[id];

	spin_lock(&bar->lock);

	bar->counter++;
	if (bar->counter == count_wait) {
	  /* last cpu to arrive */
	  spin_unlock(&bar->lock);
	  /* reset counter for next use of barrier */
	  bar->counter = 0;
	  bar->flag = bar->local_sense[id];
	}
	else {
	  spin_unlock(&bar->lock);
	  while(bar->flag != bar->local_sense[id]) ;
	}
}

#endif

#endif
