/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <barrier_param.h>
#include <sys_entry.h>
#include <list.h>
#define dbg(a,b...) //printf("\n[LIB--> %s @ %d] "a"\n",__FUNCTION__,__LINE__, ##b)

/*Create a new barrier with name `name`.
  On success returns a pointer to barrier
  On failure returns NULL
 */
barrier * barrier_open(const char *name, int oflag, int weight)
{
    struct barrier_param param;

    param.index = BARRIER_OPEN;
    param.open.name = name;
    param.open.oflag = oflag;
    param.open.weight = weight;
    param.open.ret = NULL;
    /*call _barrier_open_r*/
    nlm_sys_call(NLM_BARRIER_SYS_CALL, &param);
    return param.open.ret;
}

/*Close barrier 
  On success returns 0
  On failure returns -1
 */
int barrier_close(barrier *b)
{
    barrier_tag *barrier = NULL;
    struct barrier_param param;

    barrier = list_container_of(barrier_tag, b, weight);

    param.index = BARRIER_CLOSE;
    param.close.barrier = barrier;
    param.close.ret = 0;

    /*call _barrier_close_r*/
    nlm_sys_call(NLM_BARRIER_SYS_CALL, &param);
    return param.close.ret;
}

/*Destroy the barrier
  On success returns 0
  On failure returns -1
 */
int barrier_unlink(const char *name)
{
    struct barrier_param param;

    param.index = BARRIER_UNLINK;
    param.unlink.name = name;
    param.unlink.ret = 0;

    /*call _barrier_unlink_r*/
    nlm_sys_call(NLM_BARRIER_SYS_CALL, &param);
    return param.unlink.ret;
}

/*Initialize barrier weight
  On success returns 0
  On failure returns -1
 */
int barrier_init(barrier *b, int weight)
{
    barrier_tag *barrier = NULL;
    
    barrier = list_container_of(barrier_tag, b, weight);

    if(barrier->magic != BARRIER_MAGIC)
        return -1;

    barrier->weight = weight;
    return 0;
}

/*
   This function allows a thread to wait for a barrier to be completed. Last
   thread will return from barrier_wait with the value 1 all other threads will
   return with 0. On error -1 is returned.
 */
int barrier_wait(barrier *b)
{
    barrier_tag *barrier = NULL;
    volatile int flag; 
    barrier = list_container_of(barrier_tag, b, weight);

    dbg("b addr = %#lx, barrier = %#lx",(unsigned long)b, (unsigned long)barrier);  

    if(barrier->magic != BARRIER_MAGIC)
        return -1;

    spin_lock(&barrier->lock);
    dbg("weight = %d, cpu %d",barrier->weight,processor_id());
    flag = barrier->flag;
    if(barrier->weight == 1){
        barrier->flag = !barrier->flag;
        barrier->weight--;
        spin_unlock(&barrier->lock);
    }else{
        barrier->weight--;
        spin_unlock(&barrier->lock);
        while(flag == barrier->flag);
        return 1;
    } 
    return 0;
}
