/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <stdio.h>
#include "mipsregs.h"
#include "asmmacros.h"
#include "mips-exts.h"
#include "system.h"
#include "irq.h"
#include "cache.h"

#include "phnx_info.h"
#include "l2_cache.h"

#define read_c0_prid()          read_32bit_cp0_register($15)

static int get_processor_id() 
{
	int prid = read_c0_prid();
	return (prid >> 8) & 0xff;
}

// get number of l2 banks for this cpu
static int get_num_banks()
{
	int i;
	int prid = get_processor_id();

	for (i = 0; phnx_processors[i].name; i++) {
		if (prid == phnx_processors[i].id)
			return phnx_processors[i].banks;
	}
	printf("Unknown processor id 0x%x\n",prid);
	printf("Defaulting number of banks to 1\n");
	return 1;
}

static void write_memory (int data, uint64_t address) 
{
	volatile uint32_t *mmio0;
	mmio0 = (uint32_t *)(phoenix_io_base + 0x0);

	// Word address
	mmio0[address >> 2] = data;
}

// get l2 bank from address
static uint64_t get_bank (uint64_t address) 
{
	int num_banks = get_num_banks();
	/* get the bank mask */
	num_banks--;

	return((address & (num_banks << ADDR_OFF_BANK )) >> ADDR_OFF_BANK);
}

/* For 8 banks, default set shift is 8 bits. In case of less number of
 * banks the shift reduces. This is to calculate shift based on the number of
 * banks available in xlr/xls.
 * input should be number of banks... assumed to be 1, 2, 4 or 8
 */
static int get_bank_shift(int num_banks)
{
	int shift = 0;
	/* For num of banks = 1, set starts at bit 5, 
	 * for num of banks = 2, set starts at bit 6,
	 * for num of banks = 4, set starts at bit 7,
	 * for num of banks = 8, set starts at bit 8
	 */
	while (1) {
		if (!(num_banks >>= 1)) 
			break;
		shift++;
	}
	shift = 3 - shift;
	return shift;
}

// get l2 set from the address
static uint64_t get_set (uint64_t address) 
{
	int num_banks = get_num_banks();
	int shift = get_bank_shift(num_banks);

	return(((address & (0x3ff << (ADDR_OFF_SET-shift) )) >> 
				(ADDR_OFF_SET-shift)));
}

// flush l2 cache line as indicated by address and way 
static void flush_l2_line(uint64_t address, int way) 
{
	int l2_data;
	uint64_t l2_address;

	// Write CMD
	// data set, way, command
	l2_data = (get_set(address) << L2_CMD_OFF_SET) | 
		    (way << L2_CMD_OFF_WAY) | 
		    (L2_CMD_FLD_CMD_EVICT_LINE);

	// L2 address
	l2_address = (L2_ADDR_FLD_CONFIG_L2 << L2_ADDR_OFF_CONFIG) | 
		(L2_ADDR_FLD_REG_L2_CMD << L2_ADDR_OFF_REG) | 
		(get_bank(address) << L2_ADDR_OFF_BANK);

	write_memory(l2_data, l2_address);
}

// Flush complete l2 cache
// this flush_l2_cache is not preserving previously locked lines in the L2...
void flush_l2_cache(void)
{
	int bank=0;
	int set = 0;
	int way = 0;
	uint64_t address = 0;
	int num_banks = get_num_banks();
	int shift = get_bank_shift(num_banks);

	for(set=0;set<NUM_SETS;set++) {
		for(bank=0;bank<num_banks;bank++) {
			address = (set << (ADDR_OFF_SET-shift)) | (bank <<
					ADDR_OFF_BANK);
			for(way=0;way<NUM_WAYS;way++) {
				flush_l2_line(address, way);
			}
		}
	}	
}   

