/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sched.h>
#include <unistd.h>
#include <sys/types.h>
#include <dlfcn.h>
#include <syscall.h>
#include <errno.h>
#include <sys/shm.h>
#include <pthread.h>
#include <sys/syscall.h>
#include <nlm_hal.h>
#include "spinlock.h"

#define MAX_CPUS 128

static unsigned int cpu_online_map;
static unsigned int cpu_map = 0;
static int cpu_map_init = 0;
static int ldmode = 0, pcpu = 0;
static spinlock_t ptlock = { 0 };
struct wakeup_info
{
	pthread_t pth;
	int (*fn)(void *);
	void *arg;
};

static struct wakeup_info wakeup[MAX_CPUS];

enum {
	MODE_APP_AMP = 1,
	MODE_APP_SMP = 2
};


static int phys_to_log_map[MAX_CPUS];
static void set_phys_to_log_map(unsigned int map)
{
	int i, log_cpu = 0;
	for(i = 0; i < MAX_CPUS; i++) {
		if((1 << i) & map) {
			phys_to_log_map[i] = log_cpu;
			log_cpu++;
		}
	}
}

static int get_phys_to_log_map(int physcpu)
{
	return phys_to_log_map[physcpu];
}

static int app_set_affinity(int cpu)
{
	cpu_set_t mask;
	int tid;

	CPU_ZERO(&mask);
	CPU_SET(cpu, &mask);
	tid = syscall(SYS_gettid);
	if(sched_setaffinity(tid, sizeof(mask), &mask) != 0) {
		fprintf(stderr, "Set affinity failed for cpu %d\n", cpu);
		return -1;
	}
	return 0;
}

pid_t getpid(void)
{
	unsigned int pid = 0;

	__asm__ volatile (".set push\n"
			".set noreorder\n"
			".set arch=xlp\n"
			"rdhwr %0, $0\n"
			".set pop\n"
			: "=r" (pid)
			:);

	return (pid_t)pid;
}

static void *pthread_init(void *arg)
{
	int cpu = (int)(long)arg;
	int physcpu;
	app_set_affinity(get_phys_to_log_map(cpu));

	physcpu = getpid();
	printf("Running Child on Phys CPU %d \n", physcpu);
	if(wakeup[cpu].fn) {
		wakeup[cpu].pth = pthread_self();
		wakeup[cpu].fn(wakeup[cpu].arg);
	}
	return NULL;
}

static void ld_mode_init()
{
	char *buf;
	unsigned int cpu_possible_map;
	if(ldmode == 0) {
		if((buf = getenv("ldcpupossiblemap")) == NULL) {
			fprintf(stderr, "ldcpupossiblemap getenv failed\n");
			return;
		}
		cpu_possible_map = strtoul(buf, NULL, 16);
		set_phys_to_log_map(cpu_possible_map);

		if((buf = getenv("ldcpuonlinemap")) == NULL) {
			fprintf(stderr, "ldcpuonlinemap getenv failed\n");
			return;
		}
		cpu_online_map = strtoul(buf, NULL, 16);
		printf("%s in cpu_possible_map %x cpu_online_map %x\n", __FUNCTION__,
				cpu_possible_map, cpu_online_map);


		if((buf = getenv("ldmode")) == NULL) {
			fprintf(stderr, "ldmode getenv failed\n");
			return;
		}

		if(strcmp(buf, "amp") == 0)
			ldmode = MODE_APP_AMP;
		else if(strcmp(buf, "smp") == 0)
			ldmode = MODE_APP_SMP;
	}
	return;
}



void exit(int status)
{
	pthread_t t;
	int i;
	void (*f)();
	int cpu;
       	cpu = getpid();

	//printf("in exit cpu %d status %d\n", cpu, status);
	ld_mode_init();
	if(ldmode == MODE_APP_SMP) {
		if(!cpu_map_init)
			goto done;
		if(pcpu != cpu) {
			t = pthread_self();
			for(i = 0; i < MAX_CPUS; i++) {
				if(wakeup[i].pth && (wakeup[i].pth == t)) {
					pthread_exit(0);
				}

			}
		}

		/* Parent wait for all childs to exit only if the exit is normal */
		if(status == 0) {
			for(i = 0; i < MAX_CPUS; i++) {
				if((t = wakeup[i].pth) != 0) {
					pthread_join(t, NULL);
					wakeup[i].pth = 0;
				}
			}
		}
	}

done:
	f = dlsym(RTLD_NEXT, "exit");
	if(f != NULL) {
		f(status);
	}
	else {
		fprintf(stderr, "Unable to lookup \'exit\' symbol, aborting!\n");
	}

	abort();
}

pid_t clone(int (*fn)(void *), void *child_stack, int flags, void *arg, ...)
{
	int i;
	pthread_t child;

	ld_mode_init();
	if(ldmode != MODE_APP_SMP)
		return -1;
	if((flags & CLONE_VM) == 0)
		/* In buddy mode VM should be set as all the segments are shared. */
		return -1;

	spin_lock(&ptlock);

	if(!cpu_map_init) {

		cpu_map = cpu_online_map; 
		cpu_map_init = 1;

		/* ignore the first cpu, assuming the application is running on that */
		for(i = 0; i < sizeof(cpu_map) * 8; i++) {
			if(cpu_map & (1 << i)) {
				cpu_map &= (~(1 << i));
				pcpu = i;
				break;
			}
		}
	}

	if(cpu_map == 0) {
		spin_unlock(&ptlock);
		return -1;
	}

	for(i = 0; i < sizeof(cpu_map) * 8; i++) {
		if(cpu_map & (1 << i)) {
			cpu_map &= (~(1 << i));
			break;
		}
	}

	spin_unlock(&ptlock);

	wakeup[i].fn = fn;
       	wakeup[i].arg = arg;

	if(pthread_create(&child, NULL, (void *)pthread_init, (void *)(long)i) != 0)
		return -1;
	wakeup[i].pth = child;

	sched_yield();
	return i;
}


