/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

/**
* @file_name atomic.h
*/

/**
* @defgroup atomic Atomic API
* @brief Description about the atomic apis supported in hyperexec
*/

#define CONFIG_64BIT

#ifndef NLM_XLP_A0_WORKAROUND
#define NLM_ATOMIC 1
#endif

/**
* @brief Atomically does bitwise or operation between value pointed by p and v
* *p = *p | v
*
* @param[in] p Pointer to unsigned int.
*
* @param[in] v Unsigned int value. 
*
* @return 
*  -  void
*	  
* @ingroup atomic
*/
static __inline__ void atomic_set_int(volatile unsigned int *p, unsigned int v)
{
    unsigned int temp;
    __asm__ volatile (
            ".set push\n"
            ".set noreorder\n"
            "sync\n"
            "1:ll %0, %3\n"
            "or %0, %2\n"
            "sc %0, %1\n"
            "beqz %0, 1b\n"
            "sync\n"
            ".set pop\n"
            : "=&r"(temp), "=m"(*p)
            : "r"(v), "m"(*p)
            : "memory"
            );
}

/**
* @brief Atomically does bitwise nor operation between value pointed by p and v
* *p = *p & ~v
*
* @param[in] p Pointer to unsigned int.
*
* @param[in] v Unsigned int value. 
*
* @return 
*  -  void
*	  
* @ingroup atomic
*/
static __inline__  void atomic_clear_int(volatile unsigned int *p, 
                        unsigned int  v)
{
    unsigned int temp, temp1;
    __asm__ volatile(
            ".set push\n"
            ".set noreorder\n"
            "sync\n"
            "1: ll %0, %4\n"
            "nor %2, %3, $0\n"
            "and %0, %0, %2\n"
            "sc %0, %1\n"
            "beqz %0, 1b\n"
            "sync\n"
            ".set pop\n"
            : "=&r" (temp), "=m"(*p), "=&r"(temp1)
            : "r"(v), "m"(*p)
            : "memory");
}

#ifdef NLM_ATOMIC
static __inline__ void ldadd_w_no_read(volatile unsigned int *p, unsigned int v)
{
    __asm__ volatile(
            ".set push\n"
            ".set noreorder\n"
            "move $8, %0\n"
            "move $9, %1\n"
            "ldaddwu $8, $9\n"
            ".set pop\n"
            :
            :"r"(v), "r"((unsigned long)p)
            :"$8", "$9", "memory"
        );
}
#endif /* NLM_ATOMIC */

/**
* @brief  Atomically add v to the integer pointed by p
* *p = *p + v
*
* @param[in] p Pointer to unsigned int.
*
* @param[in] v Unsigned int value. 
*
* @return 
*  -  void
*	  
* @ingroup atomic
*/
static __inline__ void atomic_add_int(volatile unsigned int *p, unsigned int v)
{
#ifdef NLM_ATOMIC
    ldadd_w_no_read(p, v);
#else /* NLM_ATOMIC */
    unsigned int temp;

    __asm__ volatile (
            ".set push\n"
            ".set noreorder\n"
            "   sync\n"
            "1: ll   %0, %3\n"
            "   addu %0, %0, %2\n"
            "   sc   %0, %1\n"
            "   beqz %0, 1b\n"
            "    sync\n"
            ".set pop\n"
            : "=&r"(temp), "=m"(*p)
            : "r"(v), "m"(*p)
            : "memory"
            );
#endif /* NLM_ATOMIC */
}

/**
* @brief Atomically subtracts v from the integer pointed by p
* *p = *p - v
*
* @param[in] p Pointer to unsigned int.
*
* @param[in] v Unsigned int value. 
*
* @return 
*  -  void
*	  
* @ingroup atomic
*/
static __inline__ void atomic_subtract_int(volatile unsigned int *p, 
                                        unsigned int v)
{
#ifdef NLM_ATOMIC
    ldadd_w_no_read(p, -v);
#else /* NLM_ATOMIC */
    unsigned int temp;

    __asm__ volatile (
            ".set push\n"
            ".set noreorder\n"
            "   sync\n"
            "1: ll   %0, %3\n"
            "   subu %0, %0, %2\n"
            "   sc   %0, %1\n"
            "   beqz %0, 1b\n"
            "    sync\n"
            ".set pop\n"
            : "=&r"(temp), "=m"(*p)
            : "r"(v), "m"(*p)
            : "memory"
            );
#endif /* NLM_ATOMIC */
}


/**
* @brief If the value pointed by dst is exp then change it to val atomically.
*  if(*dst == exp)
*   *dst = val;
*
* @param[in] dst Pointer to unsigned int.
*
* @param[in] exp Unsigned int value. 
*
* @param[in] val Unsigned int value. 
*
* @return 
*   :On success 1
*   :On failure 0
*	  
* @ingroup atomic
*/

static __inline__ int atomic_cmpset_int(volatile unsigned int *dst, 
                                unsigned int exp, unsigned int val)
{
    int ret;
    unsigned int temp;

    __asm__ __volatile__ (
        ".set   push\n\t"
        ".set   noreorder\n\t"
        "sync\n"
        "1: ll %1, %5\n\t"
        "bne %1, %3, 2f\n\t"
        "move %0, $0\n\t"
        "move %0, %4\n\t"
        "sc %0, %2\n"
        "beqz %0, 1b\n\t"
        "sync\n"
        "2:\n"
        ".set pop\n"
        :"=&r"(ret), "=&r"(temp), "=m"(*dst)
        :"r"(exp), "r"(val), "m"(*dst)
        :"memory"
    );
    return ret;
}

/**
* @brief Atomically adds v to the integer pointed by p.
*
* @param[in] p Pointer to unsigned int.
*
* @param[in] v Unsigned int value. 
*
* @return 
*   :Returns previous value pointed by p.
*	  
* @ingroup atomic
*/

static __inline__ unsigned int atomic_fetchadd_int(volatile unsigned int *p, 
                                        unsigned int v)
{
    unsigned int old_value;
    unsigned int tmp;
    __asm__ volatile(
            ".set push\n"
            ".set noreorder\n"
            "sync\n"
            "1: ll %2, %4\n"
            "addu %1, %2, %3\n"
            "sc %1, %0\n"
            "beqz %1, 1b\n"
            "sync\n"
            ".set pop\n"
            : "=m"(*p), "=&r"(tmp), "=&r"(old_value)
            : "r"(v), "m"(*p)
            :"memory"
            );
    return old_value;
}

/**
* @brief Atomically reads the value pointed by p
*
* @param[in] p Pointer to unsigned int.
*
* @return 
*   :Returns value pointed by p.
*	  
* @ingroup atomic
*/
static __inline__ unsigned int atomic_load_acq_int(volatile unsigned int *p)
{
    unsigned int tmp;
    tmp = *p;
    __asm__ volatile("sync\n":::"memory");
    return tmp;
}

/**
* @brief Atomically sets the value pointed by p to v.
*
* @param[in] p Pointer to unsigned int.
*
* @param[in] v Unsigned int value. 
*
* @return 
*   -void
*	  
* @ingroup atomic
*/
static __inline__ void atomic_store_rel_int(volatile unsigned int *p, unsigned int v)
{
    __asm__ volatile("sync\n":::"memory");
    *p = v;
}

/**
* @brief Atomically swaps value pointed by p and v and returns the old value
* pointed by p.
*
* @param[in] p Pointer to unsigned int.
*
* @param[in] v Unsigned int value. 
*
* @return 
*   :Returns old value pointed by p.
*	  
* @ingroup atomic
*/
static __inline__ unsigned int atomic_swap_int(volatile unsigned int *p, unsigned int v)
{
#ifdef NLM_ATOMIC
    unsigned int old_val = 0 ;
    __asm__ volatile(
            ".set push\n"
            ".set noreorder\n"
            "move $8, %1\n"
            "move $9, %2\n"
            "swapwu $8, $9\n"
            "move %0, $8\n"
            ".set pop\n"
            :"=&r"(old_val)
            :"r"(v), "r"((unsigned long)p)
            :"$8","$9","memory"
            );
    return old_val;
#else /* NLM_ATOMIC */
    unsigned int old_val = 0;
    unsigned int temp;

    __asm__ volatile (
            ".set push\n"
            ".set noreorder\n"
            "   sync\n"
            "1: ll   %0, %4\n"
            "   move %2, %3\n"
            "   sc   %2, %1\n"
            "   beqz %2, 1b\n"
            "    sync\n"
            ".set pop\n"
            : "=&r"(old_val), "=m"(*p), "=&r"(temp)
            : "r"(v), "m"(*p)
            : "memory"
            );

    return old_val;
#endif /* NLM_ATOMIC */
}

#ifdef CONFIG_64BIT
/**
* @brief Atomically does bitwise or operation between value pointed by p and v
* *p = *p | v
*
* @param[in] p Pointer to unsigned long.
*
* @param[in] v Unsigned long value. 
*
* @return 
*  -  void
*	  
* @ingroup atomic
*/
static __inline__ void atomic_set_long(volatile unsigned long *p, unsigned long v)
{
    unsigned long temp;
    __asm__ volatile (
            ".set push\n"
            ".set noreorder\n"
            "sync\n"
            "1:lld %0, %3\n"
            "or %0, %2\n"
            "scd %0, %1\n"
            "beqz %0, 1b\n"
            "sync\n"
            ".set pop\n"
            : "=&r"(temp), "=m"(*p)
            : "r"(v), "m"(*p)
            : "memory"
            );
}

/**
* @brief Atomically does bitwise nor operation between value pointed by p and v
* *p = *p & ~v
*
* @param[in] p Pointer to unsigned long.
*
* @param[in] v Unsigned long value. 
*
* @return 
*  -  void
*	  
* @ingroup atomic
*/
static __inline__  void atomic_clear_long(volatile unsigned long *p, 
                        unsigned long v)
{
    unsigned long temp, temp1;
    __asm__ volatile(
            ".set push\n"
            ".set noreorder\n"
            "sync\n"
            "1: lld %0, %4\n"
            "nor %2, %3, $0\n"
            "and %0, %0, %2\n"
            "scd %0, %1\n"
            "beqz %0, 1b\n"
            "sync\n"
            ".set pop\n"
            : "=&r" (temp), "=m"(*p), "=&r"(temp1)
            : "r"(v), "m"(*p)
            : "memory");
}

#ifdef NLM_ATOMIC
static __inline__ void ldadd_d_no_read(volatile unsigned long *p, 
					unsigned long v)
{
        __asm__ volatile(
                ".set push\n"
                ".set mips64\n"
                ".set noreorder\n"
                "sync\n"
                "move $8, %0\n"
                "move $9, %1\n"
                "ldaddd $8, $9\n"
                "sync\n"
                ".set pop\n"
                :
                :"r"(v), "r"((unsigned long)p)
                :"$8","$9","memory"
                );
}
#endif /* NLM_ATOMIC */

/**
* @brief  Atomically add v to the unsigned long pointed by p
* *p = *p + v
*
* @param[in] p Pointer to unsigned long.
*
* @param[in] v Unsigned long value. 
*
* @return 
*  -  void
*	  
* @ingroup atomic
*/
static __inline__ void atomic_add_long(volatile unsigned long *p, unsigned long v)
{
#ifdef NLM_ATOMIC
    ldadd_d_no_read(p, v);
#else /* NLM_ATOMIC */
    unsigned long temp;

    __asm__ volatile (
            ".set push\n"
            ".set noreorder\n"
            "   sync\n"
            "1: lld   %0, %3\n"
            "   daddu %0, %0, %2\n"
            "   scd   %0, %1\n"
            "   beqz  %0, 1b\n"
            "    sync\n"
            ".set pop\n"
            : "=&r"(temp), "=m"(*p)
            : "r"(v), "m"(*p)
            : "memory"
            );
#endif /* NLM_ATOMIC */
}

/**
* @brief Atomically subtracts v from the long pointed by p
* *p = *p - v
*
* @param[in] p Pointer to unsigned long.
*
* @param[in] v Unsigned long value. 
*
* @return 
*  -  void
*	  
* @ingroup atomic
*/
static __inline__ void atomic_subtract_long(volatile unsigned long *p, 
                                        unsigned long v)
{
#ifdef NLM_ATOMIC
    ldadd_d_no_read(p, -v);
#else /* NLM_ATOMIC */
    unsigned long temp;

    __asm__ volatile (
            ".set push\n"
            ".set noreorder\n"
            "   sync\n"
            "1: lld   %0, %3\n"
            "   dsubu %0, %0, %2\n"
            "   scd   %0, %1\n"
            "   beqz  %0, 1b\n"
            "    sync\n"
            ".set pop\n"
            : "=&r"(temp), "=m"(*p)
            : "r"(v), "m"(*p)
            : "memory"
            );
#endif /* NLM_ATOMIC */
}

/**
* @brief If the value pointed by dst is exp then change it to val atomically.
*  if(*dst == exp)
*   *dst = val;
*
* @param[in] dst Pointer to unsigned long.
*
* @param[in] exp Unsigned long value. 
*
* @param[in] val Unsigned long value. 
*
* @return 
*   :On success 1
*   :On failure 0
*	  
* @ingroup atomic
*/
static __inline__ int atomic_cmpset_long(volatile unsigned long *dst, 
                                unsigned long exp, unsigned long val)
{
    int ret;
    unsigned long temp;

    __asm__ __volatile__ (
        ".set   push\n\t"
        ".set   noreorder\n\t"
        ".set mips64\n"
        "sync\n"
        "1: lld %1, %5\n\t"
        "bne %1, %3, 2f\n\t"
        "move %0, $0\n\t"
        "move %0, %4\n\t"
        "scd %0, %2\n"
        "beqz %0, 1b\n\t"
        "sync\n"
        "2:\n"
        ".set pop\n"
        :"=&r"(ret), "=&r"(temp), "=m"(*dst)
        :"r"(exp), "r"(val), "m"(*dst)
        :"memory"
    );
    return ret;
}

/**
* @brief Atomically adds v to the unsigned long pointed by p.
*
* @param[in] p Pointer to unsigned long.
*
* @param[in] v Unsigned long value. 
*
* @return 
*   :Returns previous value pointed by p.
*	  
* @ingroup atomic
*/
static __inline__ unsigned long atomic_fetchadd_long(volatile unsigned long *p, 
                                        unsigned long v)
{
    unsigned long old_value;
    unsigned long tmp;
    __asm__ volatile(
            ".set push\n"
            ".set noreorder\n"
            ".set mips64\n"
            "sync\n"
            "1: lld %2, %4\n"
            "daddu %1, %2, %3\n"
            "scd %1, %0\n"
            "beqz %1, 1b\n"
            "sync\n"
            ".set pop\n"
            : "=m"(*p), "=&r"(tmp), "=&r"(old_value)
            : "r"(v), "m"(*p)
            :"memory"
            );
    return old_value;
}

/**
* @brief Atomically reads the value pointed by p
*
* @param[in] p Pointer to unsigned long.
*
* @return 
*   :Returns value pointed by p.
*	  
* @ingroup atomic
*/
static __inline__ unsigned long atomic_load_acq_long(volatile unsigned long *p)
{
    unsigned long tmp;
    tmp = *p;
    __asm__ volatile("sync\n":::"memory");
    return tmp;
}

/**
* @brief Atomically sets the value pointed by p to v.
*
* @param[in] p Pointer to unsigned long.
*
* @param[in] v Unsigned long value. 
*
* @return 
*   -void
*	  
* @ingroup atomic
*/
static __inline__ void atomic_store_rel_long(volatile unsigned long *p, 
                    unsigned long v)
{
    __asm__ volatile("sync\n":::"memory");
    *p = v;
}

/**
* @brief Atomically swaps value pointed by p and v and returns the old value
* pointed by p.
*
* @param[in] p Pointer to unsigned long.
*
* @param[in] v Unsigned longvalue. 
*
* @return 
*   :Returns old value pointed by p.
*	  
* @ingroup atomic
*/
static __inline__ unsigned long atomic_swap_long(volatile unsigned long *p,
                                                unsigned long v)
{
#ifdef NLM_ATOMIC
    unsigned long old_val = 0 ;
    __asm__ volatile(
            ".set push\n"
            ".set noreorder\n"
            "move $8, %1\n"
            "move $9, %2\n"
            "swapd $8, $9\n"
            "move %0, $8\n"
            ".set pop\n"
            :"=&r"(old_val)
            :"r"(v), "r"((unsigned long)p)
            :"$8","$9","memory"
            );
    return old_val;
#else /* NLM_ATOMIC */
    unsigned long old_val = 0;
    unsigned long temp;

    __asm__ volatile (
            ".set push\n"
            ".set noreorder\n"
            "   sync\n"
            "1: lld  %0, %4\n"
            "   move %2, %3\n"
            "   scd  %2, %1\n"
            "   beqz %2, 1b\n"
            "    sync\n"
            ".set pop\n"
            : "=&r"(old_val), "=m"(*p), "=&r"(temp)
            : "r"(v), "m"(*p)
            : "memory"
            );

    return old_val;
#endif /* NLM_ATOMIC */
}
#endif
