/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "mempool_kern.h"
#include "mempool.h"
#include "contig_alloc.h"


/**
* @brief The mpcreate function creates a shared memory pool for the given name. The elements of this pool
* are allocated from a shared memory region, so that it can be accessed from all the cpus even if they
* are booted in AMP mode. This is a dynamically growing pool, initially it will have 'nminelem' elements 
* and it grows dynamically whenever a new allocation request comes. It has a per cpu cache, so that an element
* deallocated by one cpu will get back the same during the next alloc call.
*
* @param[in] nminelem Specifies the number of minimum elements allocated to this pool. Should be >= 1
* @param[in] elsize Size of each element in the pool
* @param[in] align Each element of this pool will be aligned to this value
* @param[in] flags Creation flags. Not implemented, zero now. 
*
* @return 
*  - On sucess,  returns a pointer of type mpoolptr_t. 
*  - On failure, returns NULL.
*	  
* @sa    mpalloc, mpfree, mpdestroy
*
* @ingroup mem
*/

mpoolptr_t mpcreate(const char *name, size_t nminelem, size_t elsize, size_t align, int flags)
{
	struct mempool *t;
	int i, ncpus;
	void *m;
	unsigned long long virtstart, physstart;

	if((t = contig_calloc(1, sizeof(*t))) == NULL)
		return NULL;
	strncpy(t->name, name, NAME_MAX);
	t->size = elsize;
	t->align = align;
	t->flags = flags;
	SPIN_LOCK_INIT(&t->dlock);
	list_init(&t->depot);
	
	if(!nminelem)
		goto err_exit;
	
	for(i = 0; i < MAX_CPUS; i++) {
		SPIN_LOCK_INIT(&t->clock[i]);
		list_init(&t->cache[i]);
	}

	for(i = 0; i < nminelem; i++) {
		if((m = contig_malign(align, elsize)) == NULL)
			goto err_exit;
		list_add_front(&t->depot, (struct list *)m);
		t->ndelem++;
	}

	ncpus = get_num_active_cpus();
	t->climit = nminelem / ncpus;
	t->dlimit = nminelem * 2;

	if(contig_pinfo(&virtstart, &physstart) != 0)
		goto err_exit;
	t->virtstart = (void *)(long)virtstart;
	t->physstart = physstart;
	
	return t;

err_exit:
	mpdestroy(t);
	return NULL;
	
}

/**
* @brief The mpalloc function gets an element from the memory pool
* @param[in] poolid Pool id from which the memory should be allocated. Return value from the mpcreate call. 
* @param[in] flags Alloc flags. Not implemented, zero now. 
*
* @return 
*  - On sucess,  returns a pointer to the memory object. 
*  - On failure, returns NULL.
*	  
* @sa    mpcreate, mpfree, mpdestroy
*
* @ingroup mem
*/
void *mpalloc(mpoolptr_t poolid, int flags)
{
	void *data;
	unsigned int iflags;
	struct mempool *pool = poolid;
	int my_cpu = get_my_cpu();

	/* allocate from my pool if it is not empty */
	SPIN_LOCK_IRQ_SAVE(&pool->clock[my_cpu], iflags);
	if(pool->ncelem[my_cpu]) {
		data = pool->cache[my_cpu].next;
		list_delete(data);
		pool->ncelem[my_cpu]--;
		SPIN_UNLOCK_IRQ_RESTORE(&pool->clock[my_cpu], iflags);
		return data;
	}
	SPIN_UNLOCK_IRQ_RESTORE(&pool->clock[my_cpu], iflags);

	/* allocate from depot if it is not empty */
	SPIN_LOCK_IRQ_SAVE(&pool->dlock, iflags);
	if(pool->ndelem) {
		data = pool->depot.next;
		list_delete(data);
		pool->ndelem--;
		SPIN_UNLOCK_IRQ_RESTORE(&pool->dlock, iflags);
		return data;
	}
	SPIN_UNLOCK_IRQ_RESTORE(&pool->dlock, iflags);

	/* allocate from the global pool */
	return contig_malign(pool->align, pool->size);
}

/**
* @brief The mpfree function puts back the object to the memory pool
* @param[in] poolid Pool id to which the object should be put back. 
* @param[in] obj Memory pointer returned by the mpalloc call. 
*
* @return 
*  - Returns no value.
*	  
* @sa    mpcreate, mpalloc, mpdestroy
*
* @ingroup mem
*/
void mpfree(mpoolptr_t poolid, void *obj)
{
	int my_cpu = get_my_cpu();
	unsigned int flags;
	struct mempool *pool = poolid;

	/* this should not be very accurate. so we can take the lock inside */
	if(pool->ncelem[my_cpu] < pool->climit) {
		SPIN_LOCK_IRQ_SAVE(&pool->clock[my_cpu], flags);
		list_add_front(&pool->cache[my_cpu], obj);
		pool->ncelem[my_cpu]++;
		SPIN_UNLOCK_IRQ_RESTORE(&pool->clock[my_cpu], flags);
		return;
	}

	if(pool->ndelem < pool->dlimit) {
		SPIN_LOCK_IRQ_SAVE(&pool->dlock, flags);
		list_add_front(&pool->depot, obj);
		pool->ndelem++;
		SPIN_UNLOCK_IRQ_RESTORE(&pool->dlock, flags);
		return;
	}
	contig_free(obj);
}

/**
* @brief The mpdestroy function deallocates all the elements from the memory pool
* @param[in] poolid Pool id which needs to be deallocated. 
*
* @return 
*  - Returns no value
*	  
* @sa    mpcreate, mpalloc, mpfree
*
* @ingroup mem
*/
void mpdestroy(mpoolptr_t poolid)
{
	int i;
	struct list *l, *head, *n;
	unsigned int flags;
	struct mempool *pool = poolid;

	for(i = 0; i < MAX_CPUS; i++) {
		SPIN_LOCK_IRQ_SAVE(&pool->clock[i], flags);
		head = &pool->cache[i];
		for(l = head->next; l != head; ) {
			n = l->next;
			contig_free(l);
			l = n;
		}
		SPIN_UNLOCK_IRQ_RESTORE(&pool->clock[i], flags);
	}

	SPIN_LOCK_IRQ_SAVE(&pool->dlock, flags);
	head = &pool->depot;
	for(l = head->next; l != head; ) {
		n = l->next;
		contig_free(l);
		l = n;
	}
	SPIN_UNLOCK_IRQ_RESTORE(&pool->dlock, flags);

	contig_free(pool);
}
/**
* @brief The mpgetphys function returns the physical address for the given virtual address
* @param[in] poolid Pool id from which the memory is allocated. 
* @param[in] vaddr virtual address. 
*
* @return 
*  - Returns the physical address
*	  
* @sa    mpcreate, mpalloc, mpfree
*
* @ingroup mem
*/
unsigned long long mpgetphys(mpoolptr_t poolid, void *vaddr)
{
	struct mempool *pool = poolid;
	unsigned long long off = vaddr - pool->virtstart;
	return pool->physstart + off;
}

/**
* @brief The mpgetvirt function returns the virtual address for the given physical address
* @param[in] poolid Pool id from which the memory is allocated. 
* @param[in] paddr physicall addres. 
*
* @return 
*  - Returns the virtual address
*	  
* @sa    mpcreate, mpalloc, mpfree
*
* @ingroup mem
*/
void *mpgetvirt(mpoolptr_t poolid, unsigned long long paddr)
{
	struct mempool *pool = poolid;
	unsigned long long off = paddr - pool->physstart;
	return pool->virtstart + off;
}
