/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>
#include "spinlock.h"
#include "pmemparam.h"
#include "contig_alloc.h"
#include "sys_entry.h"
#include "devices.h"
#include "tlb.h"

unsigned long devmem_vaddr_start;
struct free_vaddr_list {
	unsigned long vaddr;
	uint64_t size;
	struct free_vaddr_list *next;
};
struct alloc_vaddr_list {
	unsigned long vaddr;
	uint64_t paddr;
	uint64_t size;
	int refcnt;
	struct alloc_vaddr_list *next;
};

static struct free_vaddr_list free_vaddr_head;
static struct alloc_vaddr_list alloc_vaddr_head;
#define NTLBSZ 9
static uint32_t hpexec_pgsz;
static spinlock_t dev_mem_lock;
#define dbg_printf(fmt, args...)

static unsigned int tlb_sizes[NTLBSZ] = {
	1 << 10,        //   1 KB
	4 << 10,        //   4 KB
	16 << 10,       //  16 KB
	64 << 10,       //  64 KB
	1 << 20,        //   1 MB
	4 << 20,        //   4 MB
	16 << 20,       //  16 MB
	64 << 20,       //  64 MB
	256 << 20       // 256 MB
};

static unsigned long tlb_align_size(uint64_t size)
{
	int i;

	for (i = 4; i < NTLBSZ ; i++) {
		if(size <= (2*tlb_sizes[i]))
			return (2*tlb_sizes[i]);
	}
	return (512 << 20);
}

/* alignment */
#define   nlm_align(x, y)     ((x) & (~((y)-1)))
#define   nlm_roundup(x, y)   (nlm_align((x)+(y)-1, (y)))


static int vaddr_add(unsigned long vaddr, uint64_t paddr, uint64_t size)
{
	struct alloc_vaddr_list *tmp;
	if((tmp = malloc(sizeof(struct alloc_vaddr_list))) == NULL)
		return -1;
	tmp->refcnt = 1;
	tmp->vaddr = vaddr;
	tmp->paddr = paddr;
	tmp->size = size;
	tmp->next = alloc_vaddr_head.next;
	alloc_vaddr_head.next = tmp;
	return 0;
}

static int vaddr_remove(unsigned long *vaddr, uint64_t *size, uint64_t *paddr)
{
	struct alloc_vaddr_list *cur = alloc_vaddr_head.next;
	struct alloc_vaddr_list *prev = &alloc_vaddr_head;

	while(cur != NULL) {
		if(*vaddr >= cur->vaddr && ((*vaddr + *size) <= cur->vaddr + cur->size))
			break;
		prev = cur;
		cur = cur->next;
	}
	if(cur) {
		cur->refcnt--;
		if(paddr)
			*paddr = cur->paddr;
		if(cur->refcnt == 0) {
			*vaddr = cur->vaddr;
			*size = cur->size;
			prev->next = cur->next;
			free(cur);
			return 1;
		}
		return 0;
	}
	return -EINVAL;
}

int paddr_is_unmapped(uint64_t paddr, uint64_t size, unsigned long *rvaddr)
{
	struct alloc_vaddr_list *cur = alloc_vaddr_head.next;

	while(cur != NULL) {
		if(paddr >= cur->paddr && ((paddr + size) <= (cur->paddr + cur->size))) {
			cur->refcnt++;
			*rvaddr = cur->vaddr + (paddr - cur->paddr);
			return 0;
		}

		/* return error if there is a partial map already */
		 if(paddr >= cur->paddr && (paddr <= (cur->paddr + cur->size)))
			 return -EINVAL;

		 cur = cur->next;
	}
	return 1;			
}

int free_vaddr_init(unsigned long svaddr, unsigned long evaddr)
{
	struct free_vaddr_list *tmp;
	uint64_t size;

	devmem_vaddr_start = svaddr;
	if((tmp = malloc(sizeof(struct free_vaddr_list))) == NULL)
		return -1;
	
	/* As min_page_size is 1MB, better to have vaddr aligned to 2MB */
	hpexec_pgsz = MIN_PAGE_SIZE + MIN_PAGE_SIZE;
	svaddr = nlm_roundup(svaddr, hpexec_pgsz);
	size = evaddr - svaddr;
	dbg_printf(" %s in svaddr %lx evaddr %lx size %lx\n", __FUNCTION__, svaddr, evaddr, (long)size);

	tmp->vaddr = svaddr;
	tmp->size = size;
	tmp->next = NULL;
	free_vaddr_head.next = tmp;
	return 0;
}

static int split_mem(struct free_vaddr_list *cur, struct free_vaddr_list *prev,
		unsigned long vaddr, uint64_t size)
{
	struct free_vaddr_list *tmp;

	dbg_printf(" %s in vaddr %lx size %lx\n", __FUNCTION__, vaddr, (long)size);
	if(cur->vaddr == vaddr) {
		cur->vaddr += size;
		cur->size -= size;
		if(cur->size == 0) {
			prev->next = cur->next;
			free(cur);
		}
		return 0;
	}
	if((tmp = malloc(sizeof(struct free_vaddr_list))) == NULL)
		return -1;
	tmp->vaddr = vaddr + size;
	tmp->size = (cur->vaddr + cur->size) - tmp->vaddr;
	cur->size = vaddr - cur->vaddr;
	if(tmp->size == 0) {
		free(tmp);
		return 0;
	}
	tmp->next = cur->next;
	cur->next = tmp;
	return 0;
}

int unmap_mem(unsigned long *vaddr, uint64_t *size, uint64_t *paddr)
{
	struct free_vaddr_list *cur = free_vaddr_head.next;
	struct free_vaddr_list *prev = &free_vaddr_head;
	struct free_vaddr_list *next, *tmp;
	int rv;

	rv = vaddr_remove(vaddr, size, paddr);
	if(rv <= 0)
		return rv;

	dbg_printf(" %s in vaddr %lx size %llx\n", __FUNCTION__, *vaddr, (unsigned long long)*size);

	while(cur != NULL) {
		if(cur->vaddr > *vaddr)
			break;
		prev = cur;
		cur = cur->next;
	}

	if(cur && ((*vaddr + *size) == cur->vaddr)) {
		cur->vaddr -= *size;
		cur->size += *size;
	} else if(prev && ((prev->vaddr + prev->size) == *vaddr)) {
		prev->size += *size;
	} else {
		if((tmp = malloc(sizeof(struct free_vaddr_list))) == NULL)
			return -ENOMEM;
		tmp->size = *size;
		tmp->vaddr = *vaddr;
		prev->next = tmp;
		tmp->next = cur;
	}

	/* collapse from prev */
	cur = prev;
	while(cur != NULL) {
		if((next = cur->next) == NULL)
			break;
		if((cur->vaddr + cur->size) == next->vaddr) {
			cur->size += next->size;
			cur->next = next->next;
			free(next);
		}
		cur = cur->next;
	}
	return 1;
}


int map_mem(unsigned long addr, uint64_t paddr, uint64_t size, unsigned long *rvaddr)
{
	struct free_vaddr_list *cur = free_vaddr_head.next;
	struct free_vaddr_list *prev = &free_vaddr_head;
	unsigned long vaddr;
	uint64_t tlb_sz, reqsize;
	int rv;

	dbg_printf(" %s in size %lx\n", __FUNCTION__, (long)size);
	/* If address is specified we will align to the hpexec page size(mmap requirement), 
	   otherwise to the maximum page size to reduce the number of tlbs */
	/* return error if addr and size not aligned properly */
	if(((addr % hpexec_pgsz) != 0) || ((size % hpexec_pgsz) != 0) || ((paddr % hpexec_pgsz) != 0))
		return -EINVAL;

	/* If it is already mapped, increment the refcnt and return */
	rv = paddr_is_unmapped(paddr, size, rvaddr);
	if(rv <= 0)
		return rv;

	if(addr)
		tlb_sz = hpexec_pgsz;
	else
		tlb_sz = tlb_align_size(size);

	while(cur != NULL) {
		if(addr) {
			if((addr >= cur->vaddr) && ((addr + size) <= (cur->vaddr + cur->size))) {
				vaddr = addr;
				break;
			}
		} else {
			vaddr = nlm_roundup(cur->vaddr, tlb_sz);
			reqsize = size + vaddr - cur->vaddr;
			if(cur->size >= reqsize) 
				break;
		}
		prev = cur;
		cur = cur->next;
	}
	if(cur == NULL)  {
		/* if address is specified look for the other free vaddr */
		if(addr)
			return map_mem(0, paddr, size, rvaddr);	
		return -ENOMEM;
	}

	split_mem(cur, prev, vaddr, size);
	vaddr_add(vaddr, paddr, size);
	*rvaddr = vaddr;
	return 1;
}



/**
* @file_name pmem.c
*/

/**
* @defgroup pmem Pmem Memory Management API
* @brief Description about the memory management apis supported in NETOS
*/


static int pmem_open(struct _reent *ptr, const char *file, int flags, int mode)
{
	struct device_ops *dev;
	int pmem_close(struct _reent *reent, int fd);
	void *pmem_mmap(struct _reent *reent, int fd, void *addr, size_t len, int prot, int flags, off_t off);
	int pmem_munmap(struct _reent *reent, void *addr, size_t len);
	int fd;
	if((dev = malloc(sizeof(struct device_ops))) == NULL)
		return -1;

	if((dev->name = malloc(256)) == NULL)
		return -1;
	strncpy((char *) dev->name, "dev-mem", 256);

	dev->close_r 	= pmem_close;
	dev->mmap	= pmem_mmap;
	dev->munmap	= pmem_munmap;

	fd = fdget();
	if(dev_add(dev, fd) < 0) {
		free(dev);
		fdput(fd);
		return -1;
	}
	
	return fd;
}

void *pmem_mmap(struct _reent *reent, int fd, void *addr, size_t len, int prot, int flags, off_t off)
{

	struct netl_pmem_mmap *mmap;
	struct netl_pmemcall pmemcall;
	unsigned long vaddr;
	int rv;
	uint64_t size = len;

	/* only one call , netos uses global mem to flush the entries in munmap. */
	spin_lock(&dev_mem_lock);
	if(flags & MAP_FIXED || flags & MAP_PRIVATE) {
		reent->_errno = EINVAL;
		goto err_exit;
	}

	rv = map_mem((unsigned long)addr, off, len, &vaddr);
	if(rv < 0) {
		reent->_errno = -rv;
		goto err_exit;
	}
	/* if it is alredy mapped, */
	if(rv == 0) {
		dbg_printf("%s in cpu %d addr is already is mapped\n", __FUNCTION__, getpid());
		spin_unlock(&dev_mem_lock);
		return (void *)vaddr;
	}
	
	mmap = &pmemcall.mmap;

	pmemcall.cmd = NETL_PMEM_MMAP_CALL;
	mmap->vaddr = (uint64_t)vaddr;
	mmap->len = len;
        mmap->prot = prot;
	mmap->flags = flags;
	mmap->off = off;	
	nlm_sys_call(NLM_PMEM_SYS_CALL, (void *)&pmemcall);
	reent->_errno = pmemcall.err;

	if(pmemcall.rv != 0) {
		unmap_mem(&vaddr, &size, NULL);
		goto err_exit;
	}

	spin_unlock(&dev_mem_lock);
	return (void *)vaddr;

err_exit:
	spin_unlock(&dev_mem_lock);
	return MAP_FAILED;
};

int pmem_munmap(struct _reent *reent, void *addr, size_t len)
{
	struct netl_pmem_munmap *munmap;
	struct netl_pmemcall pmemcall;
	uint64_t paddr;
	int rv;
	unsigned long vaddr = (unsigned long)addr;
	uint64_t size = len;

	/* only one call , netos uses global mem to flush the entries in munmap. */
	spin_lock(&dev_mem_lock);
	rv = unmap_mem(&vaddr, &size, &paddr);
	 if(rv < 0) {
		reent->_errno = EINVAL;
		goto err_exit;
	}
	/* if the address is mapped by another thread */
	if(rv == 0) {
		dbg_printf("%s in cpu %d addr is already is mapped\n", __FUNCTION__, getpid());
		spin_unlock(&dev_mem_lock);
		return 0;
	}
	
	munmap = &pmemcall.munmap;

	pmemcall.cmd = NETL_PMEM_MUNMAP_CALL;
	munmap->vaddr = (uint64_t)vaddr;
	munmap->len = size;
	munmap->off = paddr;

	nlm_sys_call(NLM_PMEM_SYS_CALL, (void *)&pmemcall);
	reent->_errno = pmemcall.err;

	if(pmemcall.rv != 0)
		goto err_exit;

	spin_unlock(&dev_mem_lock);
	return 0;
err_exit:
	spin_unlock(&dev_mem_lock);
	return -EINVAL;
}

int pmem_close(struct _reent *reent, int fd)
{
	return 0;

}

static void pmem_init(void)
{

}

struct device_ops dev_mem = {
	.name 		= "/dev/mem",
	.open_r 	= pmem_open,
	.close_r 	= NULL,
	.mmap		= NULL,
	.munmap		= NULL,
	.init		= pmem_init
	
};
