/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

/**
* @file_name semapi.c
*/

/**
* @defgroup sem Semaphore API
* @brief Description about the semaphore apis supported in hyperexec
*/

#include <semaphore.h>
#include <semparam.h>
#include <stdarg.h>
#include <sys_entry.h>

#include <stdio.h>

#ifdef N32_DBG
static void print_sem_open_offsets(struct semparam *s)
{
	printf("=== HE sem open %p (%d)\n", s, sizeof(struct semparam));
        printf(" index %p\n", &s->index);
	printf(" name %p (%p:%s)\n", &s->open.name, s->open.name, s->open.name);
	printf(" oflag %p\n", &s->open.oflag);
	printf(" mode %p \n", &s->open.mode);
	printf(" value %p (%u)\n", &s->open.value, s->open.value);
	printf(" ret %p (%p) \n", &s->open.ret, s->open.ret);
}
static void print_sem_close_offsets(struct semparam *s)
{
	printf("=== HE sem close %p (%d)\n", s, sizeof(struct semparam));
        printf(" index %p\n", &s->index);
	printf(" sem 0x%llx (%p) \n", &s->close.sem, (sem_t *) (unsigned long) s->close.sem);
	printf(" ret %p (%d) \n", &s->close.ret, s->close.ret);
}
static void print_sem_unlink_offsets(struct semparam *s)
{
	printf("=== HE sem unlink %p (%d)\n", s, sizeof(struct semparam));
        printf(" index %p\n", &s->index);
	printf(" name 0x%llx (%s) \n", s->unlink.name, (char *) (unsigned long) s->unlink.name);
	printf(" ret %p (%d) \n", &s->unlink.ret, s->unlink.ret);
}
#endif
/**
* @brief Creats or opens an existing named semaphore. 
*
* @param[in] name Name of the semaphore
*
* @param[in] oflag Operation flags. Valid values are
*  - O_CREAT : Create the semaphore if it does not exist.  
*  - O_EXCL  : Used along with O_CREAT. Returns error if the semaphore object 
*              exists.
* @param[in] mode Permission to use. Currently this field is ignored. This
*                 optional parameter is used only when semaphore doesn't exist 
*                 and O_CREAT flag is passed.
* @param[in] value The semaphore is created with initial value of "value".
*                  Valid initial values for semaphores are less than or equal
*                  to {SEM_VALUE_MAX}. This optional parameter is used only 
*                  when semaphore doesn't exist and O_CREAT flag is passed.
*
* @return 
*  - On sucess,  returns address of the semaphore. 
*  - On failure, returns SEM_FAILED. 
*	  
* @sa    sem_close, sem_unlink
*
* @ingroup sem
*/
sem_t *sem_open(const char *name, int oflag, ...)
{
    struct semparam param;
    mode_t mode = 0;
    unsigned int value = 0;
    va_list ap;

   if(oflag & O_CREAT){
        va_start(ap, oflag);
        mode = va_arg(ap, mode_t);
        value = va_arg(ap, unsigned int);
        va_end(ap);
    }

    param.index = SEM_OPEN;
    param.open.name = (unsigned long) name;
    param.open.oflag = oflag;
    param.open.mode = mode;
    param.open.value = value;
    param.open.ret = 0;
    /*call _sem_open_r*/
    nlm_sys_call(NLM_SEM_SYS_CALL, (void *)&param);
 
    return (sem_t *) (unsigned long) param.open.ret;
}

/**
* @brief Close a named semaphore.
*
* @param[in] sem Pointer to semaphore
*
* @return 
*  - On sucess,  returns 0. 
*  - On failure, returns -1. 
*	  
* @sa    sem_open, sem_unlink
*
* @ingroup sem
*/
int sem_close(sem_t *sem)
{
    struct semparam param;

    param.index = SEM_CLOSE;
    param.close.sem = (unsigned long long) (int) sem;
    param.close.ret = -1;

    /*call _sem_close_r*/
    nlm_sys_call(NLM_SEM_SYS_CALL, (void *)&param);
    return param.close.ret;
}

/**
* @brief Remove a named semaphore.
*
* @param[in] name Name of the semaphore
*
* @return 
*  - On sucess,  returns 0. 
*  - On failure, returns -1. 
*	  
* @sa    sem_open, sem_close
*
* @ingroup sem
*/
int sem_unlink(const char *name)
{
    struct semparam param;

    param.index = SEM_UNLINK;
    param.unlink.name = (unsigned long long) (int) name;
    param.unlink.ret = -1;

    /*call _sem_unlink_r*/
    nlm_sys_call(NLM_SEM_SYS_CALL, (void *)&param);
    return param.unlink.ret;
}

/**
* @brief Unlock a semaphore.
*
* @param[in] sem Pointer to semaphore
*
* @return 
*  - On sucess,  returns 0. 
*  - On failure, returns -1. 
*	  
* @sa    sem_wait 
*
* @ingroup sem
*/
int sem_post(sem_t *sem)
{
    int temp;
	__asm__ __volatile__(
		"1: ll %0, %2\n"
		"addi %0, %0, 1\n"
		"sc %0, %1\n"
		"beqz %0, 1b\n"
        "nop\n"   
		: "=&r" (temp), "=m" (*sem)
		: "m" (*sem)
    );
    return 0;
}

/**
* @brief Lock a semaphore. If the semaphore value is currently zero then the
* calling thread will not return from this call.
*
* @param[in] sem Pointer to semaphore
*
* @return 
*  - On sucess,  returns 0. 
*  - On failure, returns -1. 
*	  
* @sa    sem_post, sem_trywait
*
* @ingroup sem
*/
int sem_wait(sem_t *sem)
{
	unsigned int tmp;

	__asm__ __volatile__(
        ".set push\n"
	    ".set noreorder\n"
	    "1: ll %1, %2\n"
	    "blez %1, 1b\n"
        "li $8, 1\n"
	    "sub %1, %1, $8\n"
	    "sc %1, %0\n"
	    "beqz %1, 1b\n"
	    " sync\n"
	    ".set pop\n"
	    : "=m" (*sem), "=&r" (tmp)
	    : "m" (*sem)
	    : "$8", "memory");
    return 0;
}

/**
* @brief Lock a semaphore if the semaphore value is currently not zero. This
* function doesn't block if semaphore is locked. 
*
* @param[in] sem Pointer to semaphore
*
* @return 
*  - On sucess,  returns 0. 
*  - On failure, returns -1. 
*	  
* @sa    sem_post, sem_wait
*
* @ingroup sem
*/
int sem_trywait(sem_t *sem)
{
	int tmp, ret = 0;

	__asm__ __volatile__(
        ".set push\n"
	    ".set noreorder\n"
	    "1: ll %1, %3\n"
	    "blez %1, 2f\n"
        "li %2, 1\n"
        "li $8, 1\n"
	    "sub %1, %1, $8\n"
	    "sc %1, %0\n"
	    "beqz %1, 1b\n"
        "li %2, 0\n"
	    "2: sync\n"
	    ".set pop\n"
	    : "=m" (*sem), "=&r" (tmp), "=&r"(ret)
	    : "m" (*sem)
	    : "$8", "memory");
    return -ret;
}
